@@ -29,6 +29,7 @@
 #include "media_socket.h"
 #include "rtplib.h"
 #include "ssrc.h"
+#include "codec.h"
 
 #include "rtpengine_config.h"
 
@@ -49,6 +50,7 @@ static void cli_incoming_kslist(str *instr, struct cli_writer *cw);
 static void cli_incoming_active(str *instr, struct cli_writer *cw);
 static void cli_incoming_standby(str *instr, struct cli_writer *cw);
 static void cli_incoming_debug(str *instr, struct cli_writer *cw);
+static void cli_incoming_call(str *instr, struct cli_writer *cw);
 
 static void cli_incoming_set_maxopenfiles(str *instr, struct cli_writer *cw);
 static void cli_incoming_set_maxsessions(str *instr, struct cli_writer *cw);
@@ -97,6 +99,17 @@ static void cli_incoming_list_interfaces(str *instr, struct cli_writer *cw);
 static void cli_incoming_list_jsonstats(str *instr, struct cli_writer *cw);
 static void cli_incoming_list_transcoders(str *instr, struct cli_writer *cw);
 
+static void cli_incoming_call_info(str *instr, struct cli_writer *cw);
+static void cli_incoming_call_terminate(str *instr, struct cli_writer *cw);
+static void cli_incoming_call_debug(str *instr, struct cli_writer *cw);
+static void cli_incoming_call_tag(str *instr, struct cli_writer *cw);
+
+static void cli_incoming_tag_info(str *instr, struct cli_writer *cw);
+#ifdef WITH_TRANSCODING
+static void cli_incoming_tag_delay(str *instr, struct cli_writer *cw);
+static void cli_incoming_tag_detdtmf(str *instr, struct cli_writer *cw);
+#endif
+
 static const cli_handler_t cli_top_handlers[] = {
 	{ "list",		cli_incoming_list		},
 	{ "terminate",		cli_incoming_terminate		},
@@ -109,6 +122,7 @@ static const cli_handler_t cli_top_handlers[] = {
 	{ "active",		cli_incoming_active		},
 	{ "standby",		cli_incoming_standby		},
 	{ "debug",		cli_incoming_debug		},
+	{ "call",		cli_incoming_call 		},
 	{ NULL, },
 };
 static const cli_handler_t cli_set_handlers[] = {
@@ -157,7 +171,21 @@ static const cli_handler_t cli_list_handlers[] = {
 	{ "transcoders",		cli_incoming_list_transcoders		},
 	{ NULL, },
 };
-
+static const cli_handler_t cli_call_handlers[] = {
+	{ "info",			cli_incoming_call_info			},
+	{ "terminate",			cli_incoming_call_terminate		},
+	{ "debug",			cli_incoming_call_debug			},
+	{ "tag",			cli_incoming_call_tag			},
+	{ NULL, },
+};
+static const cli_handler_t cli_tag_handlers[] = {
+	{ "info",			cli_incoming_tag_info			},
+#ifdef WITH_TRANSCODING
+	{ "delay",			cli_incoming_tag_delay			},
+	{ "detect-dtmf",		cli_incoming_tag_detdtmf		},
+#endif
+	{ NULL, },
+};
 static const cli_handler_t cli_params_handlers[] = {
 	{ "start",	cli_incoming_params_start	},
 	{ "current",	cli_incoming_params_current	},
@@ -166,6 +194,12 @@ static const cli_handler_t cli_params_handlers[] = {
 	{ NULL, },
 };
 
+
+static void cli_list_call_info(struct cli_writer *cw, struct call *c);
+static void cli_list_tag_info(struct cli_writer *cw, struct call_monologue *ml);
+
+
+
 static void cli_handler_do(const cli_handler_t *handlers, str *instr,
 		struct cli_writer *cw)
 {
@@ -289,11 +323,22 @@ static void cli_incoming_params_start(str *instr, struct cli_writer *cw) {
 			"recording-format = %s\niptables-chain = %s\n", initial_rtpe_config.b2b_url, initial_rtpe_config.redis_auth,
 			initial_rtpe_config.redis_write_auth, initial_rtpe_config.spooldir, initial_rtpe_config.rec_method,
 			initial_rtpe_config.rec_format, initial_rtpe_config.iptables_chain);
-	cw->cw_printf(cw,"listen-tcp = %s:%d\nlisten-udp = %s:%d\nlisten-ng = %s:%d\nlisten-cli = %s:%d\n",
-			sockaddr_print_buf(&initial_rtpe_config.tcp_listen_ep.address), initial_rtpe_config.tcp_listen_ep.port,
-			sockaddr_print_buf(&initial_rtpe_config.udp_listen_ep.address), initial_rtpe_config.udp_listen_ep.port,
-			sockaddr_print_buf(&initial_rtpe_config.ng_listen_ep.address), initial_rtpe_config.ng_listen_ep.port,
-			sockaddr_print_buf(&initial_rtpe_config.cli_listen_ep.address), initial_rtpe_config.cli_listen_ep.port);
+	cw->cw_printf(cw, "listen-tcp = %s\n",
+			endpoint_print_buf(&initial_rtpe_config.tcp_listen_ep[0]));
+	cw->cw_printf(cw, "listen-tcp = %s\n",
+			endpoint_print_buf(&initial_rtpe_config.tcp_listen_ep[1]));
+	cw->cw_printf(cw, "listen-udp = %s\n",
+			endpoint_print_buf(&initial_rtpe_config.udp_listen_ep[0]));
+	cw->cw_printf(cw, "listen-udp = %s\n",
+			endpoint_print_buf(&initial_rtpe_config.udp_listen_ep[1]));
+	cw->cw_printf(cw, "listen-ng = %s\n",
+			endpoint_print_buf(&initial_rtpe_config.ng_listen_ep[0]));
+	cw->cw_printf(cw, "listen-ng = %s\n",
+			endpoint_print_buf(&initial_rtpe_config.ng_listen_ep[1]));
+	cw->cw_printf(cw, "listen-cli = %s\n",
+			endpoint_print_buf(&initial_rtpe_config.cli_listen_ep[0]));
+	cw->cw_printf(cw, "listen-cli = %s\n",
+			endpoint_print_buf(&initial_rtpe_config.cli_listen_ep[1]));
 }
 
 static void cli_incoming_params_current(str *instr, struct cli_writer *cw) {
@@ -341,11 +386,22 @@ static void cli_incoming_params_current(str *instr, struct cli_writer *cw) {
 			"recording-format = %s\niptables-chain = %s\n", rtpe_config.b2b_url, rtpe_config.redis_auth,
 			rtpe_config.redis_write_auth, rtpe_config.spooldir, rtpe_config.rec_method,
 			rtpe_config.rec_format, rtpe_config.iptables_chain);
-	cw->cw_printf(cw,"listen-tcp = %s:%d\nlisten-udp = %s:%d\nlisten-ng = %s:%d\nlisten-cli = %s:%d\n",
-			sockaddr_print_buf(&rtpe_config.tcp_listen_ep.address), rtpe_config.tcp_listen_ep.port,
-			sockaddr_print_buf(&rtpe_config.udp_listen_ep.address), rtpe_config.udp_listen_ep.port,
-			sockaddr_print_buf(&rtpe_config.ng_listen_ep.address), rtpe_config.ng_listen_ep.port,
-			sockaddr_print_buf(&rtpe_config.cli_listen_ep.address), rtpe_config.cli_listen_ep.port);
+	cw->cw_printf(cw, "listen-tcp = %s\n",
+			endpoint_print_buf(&rtpe_config.tcp_listen_ep[0]));
+	cw->cw_printf(cw, "listen-tcp = %s\n",
+			endpoint_print_buf(&rtpe_config.tcp_listen_ep[1]));
+	cw->cw_printf(cw, "listen-udp = %s\n",
+			endpoint_print_buf(&rtpe_config.udp_listen_ep[0]));
+	cw->cw_printf(cw, "listen-udp = %s\n",
+			endpoint_print_buf(&rtpe_config.udp_listen_ep[1]));
+	cw->cw_printf(cw, "listen-ng = %s\n",
+			endpoint_print_buf(&rtpe_config.ng_listen_ep[0]));
+	cw->cw_printf(cw, "listen-ng = %s\n",
+			endpoint_print_buf(&rtpe_config.ng_listen_ep[1]));
+	cw->cw_printf(cw, "listen-cli = %s\n",
+			endpoint_print_buf(&rtpe_config.cli_listen_ep[0]));
+	cw->cw_printf(cw, "listen-cli = %s\n",
+			endpoint_print_buf(&rtpe_config.cli_listen_ep[1]));
 }
 
 #define int_diff_print(struct_member, option_string) \
@@ -394,30 +450,30 @@ static void cli_incoming_params_revert(str *instr, struct cli_writer *cw) {
 static void cli_incoming_list_counters(str *instr, struct cli_writer *cw) {
 	cw->cw_printf(cw, "\nCurrent per-second counters:\n\n");
 	cw->cw_printf(cw, " Packets per second (userspace)                  :%" PRIu64 "\n",
-			atomic64_get(&rtpe_stats.intv.packets_user));
+			atomic64_get(&rtpe_stats_rate.packets_user));
 	cw->cw_printf(cw, " Bytes per second (userspace)                    :%" PRIu64 "\n",
-			atomic64_get(&rtpe_stats.intv.bytes_user));
+			atomic64_get(&rtpe_stats_rate.bytes_user));
 	cw->cw_printf(cw, " Errors per second (userspace)                   :%" PRIu64 "\n",
-			atomic64_get(&rtpe_stats.intv.errors_user));
+			atomic64_get(&rtpe_stats_rate.errors_user));
 	cw->cw_printf(cw, " Packets per second (kernel)                     :%" PRIu64 "\n",
-			atomic64_get(&rtpe_stats.intv.packets_kernel));
+			atomic64_get(&rtpe_stats_rate.packets_kernel));
 	cw->cw_printf(cw, " Bytes per second (kernel)                       :%" PRIu64 "\n",
-			atomic64_get(&rtpe_stats.intv.bytes_kernel));
+			atomic64_get(&rtpe_stats_rate.bytes_kernel));
 	cw->cw_printf(cw, " Errors per second (kernel)                      :%" PRIu64 "\n",
-			atomic64_get(&rtpe_stats.intv.errors_kernel));
+			atomic64_get(&rtpe_stats_rate.errors_kernel));
 	cw->cw_printf(cw, " Packets per second (total)                      :%" PRIu64 "\n",
-			atomic64_get(&rtpe_stats.intv.packets_user) +
-			atomic64_get(&rtpe_stats.intv.packets_kernel));
+			atomic64_get(&rtpe_stats_rate.packets_user) +
+			atomic64_get(&rtpe_stats_rate.packets_kernel));
 	cw->cw_printf(cw, " Bytes per second (total)                        :%" PRIu64 "\n",
-			atomic64_get(&rtpe_stats.intv.bytes_user) +
-			atomic64_get(&rtpe_stats.intv.bytes_kernel));
+			atomic64_get(&rtpe_stats_rate.bytes_user) +
+			atomic64_get(&rtpe_stats_rate.bytes_kernel));
 	cw->cw_printf(cw, " Errors per second (total)                       :%" PRIu64 "\n",
-			atomic64_get(&rtpe_stats.intv.errors_user) +
-			atomic64_get(&rtpe_stats.intv.errors_kernel));
+			atomic64_get(&rtpe_stats_rate.errors_user) +
+			atomic64_get(&rtpe_stats_rate.errors_kernel));
 }
 
 static void cli_incoming_list_totals(str *instr, struct cli_writer *cw) {
-	AUTO_CLEANUP_INIT(GQueue *metrics, statistics_free_metrics, statistics_gather_metrics());
+	AUTO_CLEANUP_INIT(GQueue *metrics, statistics_free_metrics, statistics_gather_metrics(NULL));
 
 	for (GList *l = metrics->head; l; l = l->next) {
 		struct stats_metric *m = l->data;
@@ -537,14 +593,6 @@ static void cli_incoming_list_offertimeout(str *instr, struct cli_writer *cw) {
 
 static void cli_incoming_list_callid(str *instr, struct cli_writer *cw) {
 	struct call *c = 0;
-	struct call_monologue *ml;
-	struct call_media *md;
-	struct packet_stream *ps;
-	GList *l;
-	GList *k, *o;
-	struct timeval tim_result_duration;
-	struct timeval now;
-	char *local_addr;
 
 	if (instr->len == 0) {
 		cw->cw_printf(cw, "%s\n", "More parameters required.");
@@ -558,6 +606,17 @@ static void cli_incoming_list_callid(str *instr, struct cli_writer *cw) {
 		return;
 	}
 
+	cli_list_call_info(cw, c);
+
+	rwlock_unlock_w(&c->master_lock);	// because of call_get(..)
+	obj_put(c);
+}
+
+
+static void cli_list_call_info(struct cli_writer *cw, struct call *c) {
+	struct call_monologue *ml;
+	GList *l;
+
 	cw->cw_printf(cw,
 			 "\ncallid: %s\ndeletionmark: %s\ncreated: %i\nproxy: %s\ntos: %u\nlast_signal: %llu\n"
 			 "redis_keyspace: %i\nforeign: %s\n\n",
@@ -567,81 +626,96 @@ static void cli_incoming_list_callid(str *instr, struct cli_writer *cw) {
 
 	for (l = c->monologues.head; l; l = l->next) {
 		ml = l->data;
-		if (!ml->terminated.tv_sec)
-			gettimeofday(&now, NULL);
-		else
-			now = ml->terminated;
-
-		timeval_subtract(&tim_result_duration, &now, &ml->started);
-
-		cw->cw_printf(cw, "--- Tag '" STR_FORMAT "', type: %s, label '" STR_FORMAT "', "
-				"branch '" STR_FORMAT "', "
-				"callduration "
-				"%" TIME_T_INT_FMT ".%06" TIME_T_INT_FMT "\n",
-			STR_FMT(&ml->tag), get_tag_type_text(ml->tagtype),
-			STR_FMT(ml->label.s ? &ml->label : &STR_EMPTY),
-			STR_FMT(&ml->viabranch),
-			tim_result_duration.tv_sec,
-			tim_result_duration.tv_usec);
-
-		for (GList *sub = ml->subscriptions.head; sub; sub = sub->next) {
-			struct call_subscription *cs = sub->data;
-			struct call_monologue *csm = cs->monologue;
-			cw->cw_printf(cw, "---     subscribed to '" STR_FORMAT_M "'\n",
-					STR_FMT_M(&csm->tag));
-		}
-
-		for (k = ml->medias.head; k; k = k->next) {
-			md = k->data;
-
-			const struct rtp_payload_type *rtp_pt = __rtp_stats_codec(md);
-
-			cw->cw_printf(cw, "------ Media #%u (" STR_FORMAT " over %s) using ",
-					md->index,
-					STR_FMT(&md->type),
-					md->protocol ? md->protocol->name : "(unknown)");
-			if (!rtp_pt)
-				cw->cw_printf(cw, "unknown codec\n");
-			else
-				cw->cw_printf(cw, STR_FORMAT "\n", STR_FMT(&rtp_pt->encoding_with_params));
-
-			for (o = md->streams.head; o; o = o->next) {
-				ps = o->data;
-
-				if (PS_ISSET(ps, FALLBACK_RTCP))
-					continue;
-
-				local_addr = ps->selected_sfd ? sockaddr_print_buf(&ps->selected_sfd->socket.local.address)
-					: "0.0.0.0";
-
-				cw->cw_printf(cw, "-------- Port %15s:%-5u <> %15s:%-5u%s, SSRC %" PRIx32 ", "
-						 "" UINT64F " p, " UINT64F " b, " UINT64F " e, " UINT64F " ts",
-						 local_addr,
-						 (unsigned int) (ps->selected_sfd ? ps->selected_sfd->socket.local.port : 0),
-						 sockaddr_print_buf(&ps->endpoint.address),
-						 ps->endpoint.port,
-						 (!PS_ISSET(ps, RTP) && PS_ISSET(ps, RTCP)) ? " (RTCP)" : "",
-						 ps->ssrc_in[0] ? ps->ssrc_in[0]->parent->h.ssrc : 0,
-						 atomic64_get(&ps->stats.packets),
-						 atomic64_get(&ps->stats.bytes), atomic64_get(&ps->stats.errors),
-						 atomic64_get(&ps->last_packet));
-#if RE_HAS_MEASUREDELAY
-				if (PS_ISSET(ps, RTP) || !PS_ISSET(ps, RTCP))
-					cw->cw_printf(cw, ", %.9f delay_min, %.9f delay_avg, %.9f delay_max",
-							 (double) ps->stats.delay_min / 1000000,
-							 (double) ps->stats.delay_avg / 1000000,
-							 (double) ps->stats.delay_max / 1000000);
-#endif
-				cw->cw_printf(cw, "\n");
-			}
-		}
+		cli_list_tag_info(cw, ml);
 	}
 	cw->cw_printf(cw, "\n");
-
-	rwlock_unlock_w(&c->master_lock);	// because of call_get(..)
-	obj_put(c);
 }
 
+
+static void cli_list_tag_info(struct cli_writer *cw, struct call_monologue *ml) {
+	struct call_media *md;
+	struct packet_stream *ps;
+	struct timeval tim_result_duration;
+	struct timeval now;
+	char *local_addr;
+
+	if (!ml->terminated.tv_sec)
+		gettimeofday(&now, NULL);
+	else
+		now = ml->terminated;
+
+	timeval_subtract(&tim_result_duration, &now, &ml->started);
+
+	cw->cw_printf(cw, "--- Tag '" STR_FORMAT "', type: %s, label '" STR_FORMAT "', "
+			"branch '" STR_FORMAT "', "
+			"callduration "
+			"%" TIME_T_INT_FMT ".%06" TIME_T_INT_FMT "\n",
+		STR_FMT(&ml->tag), get_tag_type_text(ml->tagtype),
+		STR_FMT(ml->label.s ? &ml->label : &STR_EMPTY),
+		STR_FMT(&ml->viabranch),
+		tim_result_duration.tv_sec,
+		tim_result_duration.tv_usec);
+
+	for (GList *sub = ml->subscriptions.head; sub; sub = sub->next) {
+		struct call_subscription *cs = sub->data;
+		struct call_monologue *csm = cs->monologue;
+		cw->cw_printf(cw, "---     subscribed to '" STR_FORMAT_M "'\n",
+				STR_FMT_M(&csm->tag));
+	}
+	for (GList *sub = ml->subscriptions.head; sub; sub = sub->next) {
+		struct call_subscription *cs = sub->data;
+		struct call_monologue *csm = cs->monologue;
+		cw->cw_printf(cw, "---     subscription of '" STR_FORMAT_M "'\n",
+				STR_FMT_M(&csm->tag));
+	}
+
+	for (GList *k = ml->medias.head; k; k = k->next) {
+		md = k->data;
+
+		const struct rtp_payload_type *rtp_pt = __rtp_stats_codec(md);
+
+		cw->cw_printf(cw, "------ Media #%u (" STR_FORMAT " over %s) using ",
+				md->index,
+				STR_FMT(&md->type),
+				md->protocol ? md->protocol->name : "(unknown)");
+		if (!rtp_pt)
+			cw->cw_printf(cw, "unknown codec\n");
+		else
+			cw->cw_printf(cw, STR_FORMAT "\n", STR_FMT(&rtp_pt->encoding_with_params));
+
+		for (GList *o = md->streams.head; o; o = o->next) {
+			ps = o->data;
+
+			if (PS_ISSET(ps, FALLBACK_RTCP))
+				continue;
+
+			endpoint_t *local_endpoint = packet_stream_local_addr(ps);
+			local_addr = sockaddr_print_buf(&local_endpoint->address);
+
+			cw->cw_printf(cw, "-------- Port %15s:%-5u <> %15s:%-5u%s, SSRC %" PRIx32 ", "
+					 "" UINT64F " p, " UINT64F " b, " UINT64F " e, " UINT64F " ts",
+					 local_addr,
+					 (unsigned int) local_endpoint->port,
+					 sockaddr_print_buf(&ps->endpoint.address),
+					 ps->endpoint.port,
+					 (!PS_ISSET(ps, RTP) && PS_ISSET(ps, RTCP)) ? " (RTCP)" : "",
+					 ps->ssrc_in[0] ? ps->ssrc_in[0]->parent->h.ssrc : 0,
+					 atomic64_get(&ps->stats_in.packets),
+					 atomic64_get(&ps->stats_in.bytes), atomic64_get(&ps->stats_in.errors),
+					 atomic64_get(&ps->last_packet));
+#if RE_HAS_MEASUREDELAY
+			if (PS_ISSET(ps, RTP) || !PS_ISSET(ps, RTCP))
+				cw->cw_printf(cw, ", %.9f delay_min, %.9f delay_avg, %.9f delay_max",
+						 (double) ps->stats.delay_min / 1000000,
+						 (double) ps->stats.delay_avg / 1000000,
+						 (double) ps->stats.delay_max / 1000000);
+#endif
+			cw->cw_printf(cw, "\n");
+		}
+	}
+}
+
+
 static void cli_incoming_list_sessions(str *instr, struct cli_writer *cw) {
 	size_t found = 0;
 	bool all = false, own = false;
@@ -1083,6 +1157,7 @@ static void cli_incoming_standby(str *instr, struct cli_writer *cw) {
 	cli_incoming_active_standby(cw, true);
 }
 
+
 static void cli_incoming_debug(str *instr, struct cli_writer *cw) {
 	if (str_shift(instr, 1)) {
 		cw->cw_printf(cw, "No call ID specified\n");
@@ -1162,18 +1237,18 @@ static void cli_stream_readable(struct streambuf_stream *s) {
 
    free(inbuf);
    streambuf_stream_shutdown(s);
-   log_info_clear();
+   log_info_reset();
 }
 
 void cli_handle(str *instr, struct cli_writer *cw) {
 	ilogs(control, LOG_INFO, "Got CLI command: " STR_FORMAT_M, STR_FMT_M(instr));
 	cli_handler_do(cli_top_handlers, instr, cw);
+	release_closed_sockets();
 }
 
 static void cli_free(void *p) {
 	struct cli *c = p;
-	streambuf_listener_shutdown(&c->listeners[0]);
-	streambuf_listener_shutdown(&c->listeners[1]);
+	streambuf_listener_shutdown(&c->listener);
 }
 
 struct cli *cli_new(struct poller *p, endpoint_t *ep) {
@@ -1184,7 +1259,7 @@ struct cli *cli_new(struct poller *p, endpoint_t *ep) {
 
    c = obj_alloc0("cli", sizeof(*c), cli_free);
 
-   if (streambuf_listener_init(&c->listeners[0], p, ep,
+   if (streambuf_listener_init(&c->listener, p, ep,
             cli_incoming, cli_stream_readable,
             NULL,
             NULL,
@@ -1193,17 +1268,6 @@ struct cli *cli_new(struct poller *p, endpoint_t *ep) {
       ilogs(control, LOG_ERR, "Failed to open TCP control port: %s", strerror(errno));
       goto fail;
    }
-   if (ipv46_any_convert(ep)) {
-      if (streambuf_listener_init(&c->listeners[1], p, ep,
-               cli_incoming, cli_stream_readable,
-               NULL,
-               NULL,
-               &c->obj))
-      {
-         ilogs(control, LOG_ERR, "Failed to open TCP control port: %s", strerror(errno));
-         goto fail;
-      }
-   }
 
    c->poller = p;
 
@@ -1369,6 +1433,143 @@ static void cli_incoming_set_deletedelay(str *instr, struct cli_writer *cw) {
 	cw->cw_printf(cw, "Success setting delete-delay to %d\n", seconds);
 }
 
+static void cli_incoming_call(str *instr, struct cli_writer *cw) {
+	if (str_shift(instr, 1)) {
+		cw->cw_printf(cw, "More parameters required.\n");
+		return;
+	}
+
+	str callid;
+	if (str_token_sep(&callid, instr, ' '))
+		callid = STR_NULL;
+
+	if (!callid.len) {
+		cw->cw_printf(cw, "No call ID specified\n");
+		return;
+	}
+
+	cw->call = call_get(&callid);
+	if (!cw->call) {
+		cw->cw_printf(cw, "No such call '" STR_FORMAT "'\n", STR_FMT(&callid));
+		return;
+	}
+
+	cli_handler_do(cli_call_handlers, instr, cw);
+
+	if (cw->call) {
+		rwlock_unlock_w(&cw->call->master_lock);
+		obj_put(cw->call);
+		cw->call = NULL;
+	}
+}
+
+
+
+static void cli_incoming_call_info(str *instr, struct cli_writer *cw) {
+	cli_list_call_info(cw, cw->call);
+}
+static void cli_incoming_call_terminate(str *instr, struct cli_writer *cw) {
+	cw->cw_printf(cw, "\nCall '" STR_FORMAT "' terminated.\n\n", STR_FMT(&cw->call->callid));
+	ilog(LOG_WARN, "Call " STR_FORMAT_M " terminated by operator", STR_FMT_M(&cw->call->callid));
+	rwlock_unlock_w(&cw->call->master_lock);
+	call_destroy(cw->call);
+	obj_put(cw->call);
+	cw->call = NULL;
+}
+static void cli_incoming_call_debug(str *instr, struct cli_writer *cw) {
+	str_shift(instr, 1);
+
+	int flag = 1;
+
+	if (instr->len) {
+		if (!str_cmp(instr, "on") || !str_cmp(instr, "enable"))
+			;
+		else if (!str_cmp(instr, "off") || !str_cmp(instr, "disable"))
+			flag = 0;
+		else {
+			cw->cw_printf(cw, "Invalid on/off flag ('" STR_FORMAT "') specified\n", STR_FMT(instr));
+			return;
+		}
+	}
+
+	cw->call->debug = flag;
+
+	cw->cw_printf(cw, "%s debugging for call '" STR_FORMAT "'\n", flag ? "Enabled" : "Disabled",
+			STR_FMT(&cw->call->callid));
+}
+static void cli_incoming_call_tag(str *instr, struct cli_writer *cw) {
+	if (str_shift(instr, 1)) {
+		cw->cw_printf(cw, "More parameters required.\n");
+		return;
+	}
+
+	str tag;
+	if (str_token_sep(&tag, instr, ' '))
+		tag = STR_NULL;
+
+	if (!tag.len) {
+		cw->cw_printf(cw, "No tag specified\n");
+		return;
+	}
+
+	cw->ml = call_get_monologue(cw->call, &tag);
+	if (!cw->ml) {
+		cw->cw_printf(cw, "No such tag '" STR_FORMAT "'\n", STR_FMT(&tag));
+		return;
+	}
+
+	cli_handler_do(cli_tag_handlers, instr, cw);
+
+	cw->ml = NULL;
+}
+
+
+
+static void cli_incoming_tag_info(str *instr, struct cli_writer *cw) {
+	cli_list_tag_info(cw, cw->ml);
+}
+
+#ifdef WITH_TRANSCODING
+static void cli_incoming_tag_delay(str *instr, struct cli_writer *cw) {
+	if (str_shift(instr, 1)) {
+		cw->cw_printf(cw, "More parameters required.\n");
+		return;
+	}
+
+	int delay = str_to_i(instr, 1);
+	if (delay < 0) {
+		cw->cw_printf(cw, "Invalid delay %i\n", delay);
+		return;
+	}
+
+	cw->cw_printf(cw, "Setting delay to %i\n", delay);
+
+	for (GList *l = cw->ml->medias.head; l; l = l->next) {
+		struct call_media *m = l->data;
+		m->buffer_delay = delay;
+	}
+	codec_update_all_handlers(cw->ml);
+}
+static void cli_incoming_tag_detdtmf(str *instr, struct cli_writer *cw) {
+	if (str_shift(instr, 1)) {
+		cw->cw_printf(cw, "More parameters required.\n");
+		return;
+	}
+
+	int onoff = str_to_i(instr, 1);
+	if (onoff != 0 && onoff != 1) {
+		cw->cw_printf(cw, "Invalid setting %i\n", onoff);
+		return;
+	}
+
+	cw->cw_printf(cw, "%s audio DTMF detection\n", onoff ? "Enabling" : "Disabling");
+
+	cw->ml->detect_dtmf = onoff ? 1 : 0;
+	codec_update_all_handlers(cw->ml);
+}
+#endif
+
+
 
 static void cli_incoming_list_rediscmdtimeout(str *instr, struct cli_writer *cw) {
 	rwlock_lock_r(&rtpe_config.config_lock);
@@ -1460,7 +1661,7 @@ static void cli_incoming_list_interfaces(str *instr, struct cli_writer *cw) {
 }
 
 static void cli_incoming_list_jsonstats(str *instr, struct cli_writer *cw) {
-	AUTO_CLEANUP_INIT(GQueue *metrics, statistics_free_metrics, statistics_gather_metrics());
+	AUTO_CLEANUP_INIT(GQueue *metrics, statistics_free_metrics, statistics_gather_metrics(NULL));
 
 	for (GList *l = metrics->head; l; l = l->next) {
 		struct stats_metric *m = l->data;
@@ -1514,7 +1715,6 @@ static void cli_incoming_list_controltos(str *instr, struct cli_writer *cw) {
 static void cli_incoming_set_controltos(str *instr, struct cli_writer *cw) {
 	long tos;
 	char *endptr;
-	int i;
 
 	if (str_shift(instr, 1)) {
 		cw->cw_printf(cw, "%s\n", "More parameters required.");
@@ -1532,9 +1732,9 @@ static void cli_incoming_set_controltos(str *instr, struct cli_writer *cw) {
 	rtpe_config.control_tos = tos;
 	rwlock_unlock_w(&rtpe_config.config_lock);
 
-	for (i=0; i < G_N_ELEMENTS(rtpe_control_ng->udp_listeners); i++) {
-		if (rtpe_control_ng->udp_listeners[i].fd != -1) {
-			set_tos(&rtpe_control_ng->udp_listeners[i],tos);
+	for (int i = 0; i < G_N_ELEMENTS(rtpe_control_ng); i++) {
+		if (rtpe_control_ng[i]->udp_listener.fd != -1) {
+			set_tos(&rtpe_control_ng[i]->udp_listener, tos);
 		}
 	}
 
