@@ -15,30 +15,31 @@
 #include "rtplib.h"
 #include "rtcplib.h"
 #include "main.h"
+#include "ssllib.h"
 
 
 
 
-GString __thread *crypto_debug_string;
+__thread GString *crypto_debug_string;
 
 
 
-static int aes_cm_encrypt_rtp(struct crypto_context *, struct rtp_header *, str *, u_int64_t);
-static int aes_cm_encrypt_rtcp(struct crypto_context *, struct rtcp_packet *, str *, u_int64_t);
-static int aes_gcm_encrypt_rtp(struct crypto_context *, struct rtp_header *, str *, u_int64_t);
-static int aes_gcm_decrypt_rtp(struct crypto_context *, struct rtp_header *, str *, u_int64_t);
-static int aes_gcm_encrypt_rtcp(struct crypto_context *, struct rtcp_packet *, str *, u_int64_t);
-static int aes_gcm_decrypt_rtcp(struct crypto_context *, struct rtcp_packet *, str *, u_int64_t);
-static int hmac_sha1_rtp(struct crypto_context *, char *out, str *in, u_int64_t);
+static int aes_cm_encrypt_rtp(struct crypto_context *, struct rtp_header *, str *, uint64_t);
+static int aes_cm_encrypt_rtcp(struct crypto_context *, struct rtcp_packet *, str *, uint64_t);
+static int aes_gcm_encrypt_rtp(struct crypto_context *, struct rtp_header *, str *, uint64_t);
+static int aes_gcm_decrypt_rtp(struct crypto_context *, struct rtp_header *, str *, uint64_t);
+static int aes_gcm_encrypt_rtcp(struct crypto_context *, struct rtcp_packet *, str *, uint64_t);
+static int aes_gcm_decrypt_rtcp(struct crypto_context *, struct rtcp_packet *, str *, uint64_t);
+static int hmac_sha1_rtp(struct crypto_context *, char *out, str *in, uint64_t);
 static int hmac_sha1_rtcp(struct crypto_context *, char *out, str *in);
-static int aes_f8_encrypt_rtp(struct crypto_context *c, struct rtp_header *r, str *s, u_int64_t idx);
-static int aes_f8_encrypt_rtcp(struct crypto_context *c, struct rtcp_packet *r, str *s, u_int64_t idx);
+static int aes_f8_encrypt_rtp(struct crypto_context *c, struct rtp_header *r, str *s, uint64_t idx);
+static int aes_f8_encrypt_rtcp(struct crypto_context *c, struct rtcp_packet *r, str *s, uint64_t idx);
 static int aes_cm_session_key_init(struct crypto_context *c);
 static int aes_gcm_session_key_init(struct crypto_context *c);
 static int aes_f8_session_key_init(struct crypto_context *c);
 static int evp_session_key_cleanup(struct crypto_context *c);
-static int null_crypt_rtp(struct crypto_context *c, struct rtp_header *r, str *s, u_int64_t idx);
-static int null_crypt_rtcp(struct crypto_context *c, struct rtcp_packet *r, str *s, u_int64_t idx);
+static int null_crypt_rtp(struct crypto_context *c, struct rtp_header *r, str *s, uint64_t idx);
+static int null_crypt_rtcp(struct crypto_context *c, struct rtcp_packet *r, str *s, uint64_t idx);
 
 /* all lengths are in bytes */
 struct crypto_suite __crypto_suites[] = {
@@ -328,13 +329,14 @@ struct crypto_suite __crypto_suites[] = {
 	},
 };
 
+/* those crypto suites we can */
 const struct crypto_suite *crypto_suites = __crypto_suites;
 const unsigned int num_crypto_suites = G_N_ELEMENTS(__crypto_suites);
 
 
 
 
-const struct crypto_suite *crypto_find_suite(const str *s) {
+const struct crypto_suite * crypto_find_suite(const str *s) {
 	int i, l;
 	const struct crypto_suite *cs;
 
@@ -356,8 +358,6 @@ const struct crypto_suite *crypto_find_suite(const str *s) {
 	return NULL;
 }
 
-
-
 /* rfc 3711 section 4.1 and 4.1.1
  * "in" and "out" MAY point to the same buffer */
 static void aes_ctr(unsigned char *out, str *in, EVP_CIPHER_CTX *ecc, const unsigned char *iv) {
@@ -366,7 +366,8 @@ static void aes_ctr(unsigned char *out, str *in, EVP_CIPHER_CTX *ecc, const unsi
 	unsigned char *p, *q;
 	unsigned int left;
 	int outlen, i;
-	u_int64_t *pi, *qi, *ki;
+	gboolean aligned = TRUE;
+	uint64_t *pi, *qi, *ki;
 
 	if (!ecc)
 		return;
@@ -377,11 +378,16 @@ static void aes_ctr(unsigned char *out, str *in, EVP_CIPHER_CTX *ecc, const unsi
 	ki = (void *) key_block;
 	left = in->len;
 
+	if ((GPOINTER_TO_UINT(pi) % sizeof(*pi)) != 0)
+		aligned = FALSE;
+	if ((GPOINTER_TO_UINT(qi) % sizeof(*qi)) != 0)
+		aligned = FALSE;
+
 	while (left) {
 		EVP_EncryptUpdate(ecc, key_block, &outlen, ivx, 16);
 		assert(outlen == 16);
 
-		if (G_UNLIKELY(left < 16)) {
+		if (left < 16 || !aligned) {
 			p = (void *) pi;
 			q = (void *) qi;
 			for (i = 0; i < 16; i++) {
@@ -390,12 +396,14 @@ static void aes_ctr(unsigned char *out, str *in, EVP_CIPHER_CTX *ecc, const unsi
 				if (!left)
 					goto done;
 			}
-			abort();
 		}
-
-		*qi++ = *pi++ ^ ki[0];
-		*qi++ = *pi++ ^ ki[1];
-		left -= 16;
+		else {
+			qi[0] = pi[0] ^ ki[0];
+			qi[1] = pi[1] ^ ki[1];
+			left -= 16;
+		}
+		qi += 2;
+		pi += 2;
 
 		for (i = 15; i >= 0; i--) {
 			ivx[i]++;
@@ -487,7 +495,7 @@ int crypto_gen_session_key(struct crypto_context *c, str *out, unsigned char lab
 			"%02x%02x%02x%02x..., "
 			"master salt "
 			"%02x%02x%02x%02x..., "
-			"label %02x, length %i, result "
+			"label %02x, length %zu, result "
 			"%02x%02x%02x%02x...",
 			c->params.master_key[0],
 			c->params.master_key[1],
@@ -519,10 +527,10 @@ int crypto_gen_session_key(struct crypto_context *c, str *out, unsigned char lab
  */
 
 /* rfc 3711 section 4.1.1 */
-static int aes_cm_encrypt(struct crypto_context *c, u_int32_t ssrc, str *s, u_int64_t idx) {
+static int aes_cm_encrypt(struct crypto_context *c, uint32_t ssrc, str *s, uint64_t idx) {
 	unsigned char iv[16];
-	u_int32_t *ivi;
-	u_int32_t idxh, idxl;
+	uint32_t *ivi;
+	uint32_t idxh, idxl;
 
 	memcpy(iv, c->session_salt, 14);
 	iv[14] = iv[15] = '\0';
@@ -541,29 +549,42 @@ static int aes_cm_encrypt(struct crypto_context *c, u_int32_t ssrc, str *s, u_in
 }
 
 /* rfc 3711 section 4.1 */
-static int aes_cm_encrypt_rtp(struct crypto_context *c, struct rtp_header *r, str *s, u_int64_t idx) {
+static int aes_cm_encrypt_rtp(struct crypto_context *c, struct rtp_header *r, str *s, uint64_t idx) {
 	return aes_cm_encrypt(c, r->ssrc, s, idx);
 }
 
 /* rfc 3711 sections 3.4 and 4.1 */
-static int aes_cm_encrypt_rtcp(struct crypto_context *c, struct rtcp_packet *r, str *s, u_int64_t idx) {
+static int aes_cm_encrypt_rtcp(struct crypto_context *c, struct rtcp_packet *r, str *s, uint64_t idx) {
 	return aes_cm_encrypt(c, r->ssrc, s, idx);
 }
 
 /* rfc 7714 section 8 */
 
-static int aes_gcm_encrypt_rtp(struct crypto_context *c, struct rtp_header *r, str *s, u_int64_t idx) {
-	unsigned char iv[12];
+union aes_gcm_rtp_iv {
+	unsigned char bytes[12];
+	struct {
+		uint16_t zeros;
+		uint32_t ssrc;
+		uint32_t roq;
+		uint16_t seq;
+	} __attribute__((__packed__));
+} __attribute__((__packed__));
+
+_Static_assert(offsetof(union aes_gcm_rtp_iv, seq) == 10,
+               "union aes_gcm_rtp_iv not packed");
+
+static int aes_gcm_encrypt_rtp(struct crypto_context *c, struct rtp_header *r, str *s, uint64_t idx) {
+	union aes_gcm_rtp_iv iv;
 	int len, ciphertext_len;
 
-	memcpy(iv, c->session_salt, 12);
+	memcpy(iv.bytes, c->session_salt, 12);
 
-	*(u_int32_t*)(iv+2) ^= r->ssrc;
-	*(u_int32_t*)(iv+6) ^= htonl((idx & 0x00ffffffff0000ULL) >> 16);
-	*(u_int16_t*)(iv+10) ^= htons(idx & 0x00ffffULL);
+	iv.ssrc ^= r->ssrc;
+	iv.roq ^= htonl((idx & 0x00ffffffff0000ULL) >> 16);
+	iv.seq ^= htons(idx & 0x00ffffULL);
 
 	EVP_EncryptInit_ex(c->session_key_ctx[0], c->params.crypto_suite->aead_evp(), NULL,
-			(const unsigned char *) c->session_key, iv);
+			(const unsigned char *) c->session_key, iv.bytes);
 
 	// nominally 12 bytes of AAD
 	EVP_EncryptUpdate(c->session_key_ctx[0], NULL, &len, (void *)r, s->s - (char *)r);
@@ -581,18 +602,21 @@ static int aes_gcm_encrypt_rtp(struct crypto_context *c, struct rtp_header *r, s
 	return 0;
 }
 
-static int aes_gcm_decrypt_rtp(struct crypto_context *c, struct rtp_header *r, str *s, u_int64_t idx) {
-	unsigned char iv[12];
+static int aes_gcm_decrypt_rtp(struct crypto_context *c, struct rtp_header *r, str *s, uint64_t idx) {
+	union aes_gcm_rtp_iv iv;
 	int len, plaintext_len;
 
-	memcpy(iv, c->session_salt, 12);
+	if (s->len < 16)
+		return -1;
 
-	*(u_int32_t*)(iv+2) ^= r->ssrc;
-	*(u_int32_t*)(iv+6) ^= htonl((idx & 0x00ffffffff0000ULL) >> 16);
-	*(u_int16_t*)(iv+10) ^= htons(idx & 0x00ffffULL);
+	memcpy(iv.bytes, c->session_salt, 12);
+
+	iv.ssrc ^= r->ssrc;
+	iv.roq ^= htonl((idx & 0x00ffffffff0000ULL) >> 16);
+	iv.seq ^= htons(idx & 0x00ffffULL);
 
 	EVP_DecryptInit_ex(c->session_key_ctx[0], c->params.crypto_suite->aead_evp(), NULL,
-			(const unsigned char *) c->session_key, iv);
+			(const unsigned char *) c->session_key, iv.bytes);
 
 	// nominally 12 bytes of AAD
 	EVP_DecryptUpdate(c->session_key_ctx[0], NULL, &len, (void *)r, s->s - (char *)r);
@@ -612,23 +636,36 @@ static int aes_gcm_decrypt_rtp(struct crypto_context *c, struct rtp_header *r, s
 
 /* rfc 7714 section 9 */
 
-static int aes_gcm_encrypt_rtcp(struct crypto_context *c, struct rtcp_packet *r, str *s, u_int64_t idx) {
-	unsigned char iv[12];
-	unsigned char e_idx[4];
+union aes_gcm_rtcp_iv {
+	unsigned char bytes[12];
+	struct {
+		uint16_t zeros_a;
+		uint32_t ssrc;
+		uint16_t zeros_b;
+		uint32_t srtcp;
+	} __attribute__((__packed__));
+} __attribute__((__packed__));
+
+_Static_assert(offsetof(union aes_gcm_rtcp_iv, srtcp) == 8,
+               "union aes_gcm_rtcp_iv not packed");
+
+static int aes_gcm_encrypt_rtcp(struct crypto_context *c, struct rtcp_packet *r, str *s, uint64_t idx) {
+	union aes_gcm_rtcp_iv iv;
+	uint32_t e_idx;
 	int len, ciphertext_len;
 
-	memcpy(iv, c->session_salt, 12);
+	memcpy(iv.bytes, c->session_salt, 12);
 
-	*(u_int32_t*)(iv+2) ^= r->ssrc;
-	*(u_int32_t*)(iv+8) ^= htonl(idx & 0x007fffffffULL);
-	*(u_int32_t*)e_idx = htonl( (idx&0x007fffffffULL) | 0x80000000);
+	iv.ssrc ^= r->ssrc;
+	iv.srtcp ^= htonl(idx & 0x007fffffffULL);
+	e_idx = htonl( (idx&0x007fffffffULL) | 0x80000000);
 
 	EVP_EncryptInit_ex(c->session_key_ctx[0], c->params.crypto_suite->aead_evp(), NULL,
-			(const unsigned char *) c->session_key, iv);
+			(const unsigned char *) c->session_key, iv.bytes);
 
 	// nominally 8 + 4 bytes of AAD
 	EVP_EncryptUpdate(c->session_key_ctx[0], NULL, &len, (void *)r, s->s - (char *)r);
-	EVP_EncryptUpdate(c->session_key_ctx[0], NULL, &len, (void *)e_idx, 4);
+	EVP_EncryptUpdate(c->session_key_ctx[0], NULL, &len, (void *)&e_idx, 4);
 
 	EVP_EncryptUpdate(c->session_key_ctx[0], (unsigned char *) s->s, &len,
 			(const unsigned char *) s->s, s->len);
@@ -643,23 +680,23 @@ static int aes_gcm_encrypt_rtcp(struct crypto_context *c, struct rtcp_packet *r,
 	return 0;
 }
 
-static int aes_gcm_decrypt_rtcp(struct crypto_context *c, struct rtcp_packet *r, str *s, u_int64_t idx) {
-	unsigned char iv[12];
-	unsigned char e_idx[4];
+static int aes_gcm_decrypt_rtcp(struct crypto_context *c, struct rtcp_packet *r, str *s, uint64_t idx) {
+	union aes_gcm_rtcp_iv iv;
+	uint32_t e_idx;
 	int len, plaintext_len;
 
-	memcpy(iv, c->session_salt, 12);
+	memcpy(iv.bytes, c->session_salt, 12);
 
-	*(u_int32_t*)(iv+2) ^= r->ssrc;
-	*(u_int32_t*)(iv+8) ^= htonl(idx & 0x007fffffffULL);
-	*(u_int32_t*)e_idx = htonl( (idx&0x007fffffffULL) | 0x80000000);
+	iv.ssrc ^= r->ssrc;
+	iv.srtcp ^= htonl(idx & 0x007fffffffULL);
+	e_idx = htonl( (idx&0x007fffffffULL) | 0x80000000);
 
 	EVP_DecryptInit_ex(c->session_key_ctx[0], c->params.crypto_suite->aead_evp(), NULL,
-			(const unsigned char *) c->session_key, iv);
+			(const unsigned char *) c->session_key, iv.bytes);
 
 	// nominally 8 + 4 bytes of AAD
 	EVP_DecryptUpdate(c->session_key_ctx[0], NULL, &len, (void *)r, s->s - (char *)r);
-	EVP_DecryptUpdate(c->session_key_ctx[0], NULL, &len, (void *)e_idx, 4);
+	EVP_DecryptUpdate(c->session_key_ctx[0], NULL, &len, (void *)&e_idx, 4);
 
 	// decrypt partial buffer - the last 16 bytes are the tag
 	EVP_DecryptUpdate(c->session_key_ctx[0], (unsigned char *) s->s, &len,
@@ -681,10 +718,10 @@ static void aes_128_f8_encrypt(struct crypto_context *c, unsigned char *iv, str
 		      ivx[16], /* IV' */
 		      x[16];
 	int i, outlen, left;
-	u_int32_t j;
+	uint32_t j;
 	unsigned char *p;
-	u_int64_t *pi, *ki, *lki, *xi;
-	u_int32_t *xu;
+	uint64_t *pi, *ki, *lki, *xi;
+	uint32_t *xu;
 
 	EVP_EncryptUpdate(c->session_key_ctx[1], ivx, &outlen, iv, 16);
 	assert(outlen == 16);
@@ -736,9 +773,9 @@ done:
 }
 
 /* rfc 3711 section 4.1.2.2 */
-static int aes_f8_encrypt_rtp(struct crypto_context *c, struct rtp_header *r, str *s, u_int64_t idx) {
+static int aes_f8_encrypt_rtp(struct crypto_context *c, struct rtp_header *r, str *s, uint64_t idx) {
 	unsigned char iv[16];
-	u_int32_t roc;
+	uint32_t roc;
 
 	iv[0] = '\0';
 	memcpy(&iv[1], &r->m_pt, 11); /* m, pt, seq, ts, ssrc */
@@ -751,9 +788,9 @@ static int aes_f8_encrypt_rtp(struct crypto_context *c, struct rtp_header *r, st
 }
 
 /* rfc 3711 section 4.1.2.3 */
-static int aes_f8_encrypt_rtcp(struct crypto_context *c, struct rtcp_packet *r, str *s, u_int64_t idx) {
+static int aes_f8_encrypt_rtcp(struct crypto_context *c, struct rtcp_packet *r, str *s, uint64_t idx) {
 	unsigned char iv[16];
-	u_int32_t i;
+	uint32_t i;
 
 	memset(iv, 0, 4);
 	i = htonl(0x80000000ULL | idx);
@@ -765,9 +802,24 @@ static int aes_f8_encrypt_rtcp(struct crypto_context *c, struct rtcp_packet *r,
 	return 0;
 }
 /* rfc 3711, sections 4.2 and 4.2.1 */
-static int hmac_sha1_rtp(struct crypto_context *c, char *out, str *in, u_int64_t index) {
+static int hmac_sha1_rtp(struct crypto_context *c, char *out, str *in, uint64_t index) {
 	unsigned char hmac[20];
-	u_int32_t roc;
+	uint32_t roc;
+
+	roc = htonl((index & 0xffffffff0000ULL) >> 16);
+
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+	EVP_MAC_CTX *hc;
+
+	hc = EVP_MAC_CTX_dup(rtpe_hmac_sha1_base);
+	EVP_MAC_init(hc, (unsigned char *) c->session_auth_key,
+			c->params.crypto_suite->srtp_auth_key_len, NULL);
+	EVP_MAC_update(hc, (unsigned char *) in->s, in->len);
+	EVP_MAC_update(hc, (unsigned char *) &roc, sizeof(roc));
+	size_t outsize = sizeof(hmac);
+	EVP_MAC_final(hc, hmac, &outsize, outsize);
+	EVP_MAC_CTX_free(hc);
+#else // <3.0
 	HMAC_CTX *hc;
 
 #if OPENSSL_VERSION_NUMBER >= 0x10100000L
@@ -780,13 +832,13 @@ static int hmac_sha1_rtp(struct crypto_context *c, char *out, str *in, u_int64_t
 
 	HMAC_Init_ex(hc, c->session_auth_key, c->params.crypto_suite->srtp_auth_key_len, EVP_sha1(), NULL);
 	HMAC_Update(hc, (unsigned char *) in->s, in->len);
-	roc = htonl((index & 0xffffffff0000ULL) >> 16);
 	HMAC_Update(hc, (unsigned char *) &roc, sizeof(roc));
 	HMAC_Final(hc, hmac, NULL);
 #if OPENSSL_VERSION_NUMBER >= 0x10100000L
 	HMAC_CTX_free(hc);
 #else
 	HMAC_CTX_cleanup(hc);
+#endif
 #endif
 
 	assert(sizeof(hmac) >= c->params.crypto_suite->srtp_auth_tag);
@@ -799,8 +851,12 @@ static int hmac_sha1_rtp(struct crypto_context *c, char *out, str *in, u_int64_t
 static int hmac_sha1_rtcp(struct crypto_context *c, char *out, str *in) {
 	unsigned char hmac[20];
 
-	HMAC(EVP_sha1(), c->session_auth_key, c->params.crypto_suite->srtcp_auth_key_len,
-			(unsigned char *) in->s, in->len, hmac, NULL);
+	if (!HMAC(EVP_sha1(), c->session_auth_key, c->params.crypto_suite->srtcp_auth_key_len,
+			(unsigned char *) in->s, in->len, hmac, NULL))
+	{
+		memset(out, 0, c->params.crypto_suite->srtcp_auth_tag);
+		return 1;
+	}
 
 	assert(sizeof(hmac) >= c->params.crypto_suite->srtcp_auth_tag);
 	memcpy(out, hmac, c->params.crypto_suite->srtcp_auth_tag);
@@ -886,10 +942,10 @@ static int evp_session_key_cleanup(struct crypto_context *c) {
 	return 0;
 }
 
-static int null_crypt_rtp(struct crypto_context *c, struct rtp_header *r, str *s, u_int64_t idx) {
+static int null_crypt_rtp(struct crypto_context *c, struct rtp_header *r, str *s, uint64_t idx) {
 	return 0;
 }
-static int null_crypt_rtcp(struct crypto_context *c, struct rtcp_packet *r, str *s, u_int64_t idx) {
+static int null_crypt_rtcp(struct crypto_context *c, struct rtcp_packet *r, str *s, uint64_t idx) {
 	return 0;
 }
 
