@@ -14,6 +14,7 @@
 
 
 __thread struct log_info log_info;
+__thread GSList *log_info_stack;
 
 int _log_facility_cdr = 0;
 int _log_facility_rtcp = 0;
@@ -43,10 +44,18 @@ static void ilog_prefix_default(char *prefix, size_t prefix_len) {
 					STR_FMT_M(&log_info.u.call->callid));
 			break;
 		case LOG_INFO_STREAM_FD:
-			if (log_info.u.stream_fd->call)
-				snprintf(prefix, prefix_len, "[" STR_FORMAT_M " port %5u]: ",
-						STR_FMT_M(&log_info.u.stream_fd->call->callid),
-						log_info.u.stream_fd->socket.local.port);
+			if (log_info.u.stream_fd->call) {
+				if (log_info.u.stream_fd->stream)
+					snprintf(prefix, prefix_len, "[" STR_FORMAT_M "/" STR_FORMAT_M "/%u port %5u]: ",
+							STR_FMT_M(&log_info.u.stream_fd->call->callid),
+							STR_FMT_M(&log_info.u.stream_fd->stream->media->monologue->tag),
+							log_info.u.stream_fd->stream->media->index,
+							log_info.u.stream_fd->socket.local.port);
+				else
+					snprintf(prefix, prefix_len, "[" STR_FORMAT_M " port %5u]: ",
+							STR_FMT_M(&log_info.u.stream_fd->call->callid),
+							log_info.u.stream_fd->socket.local.port);
+			}
 			else
 				snprintf(prefix, prefix_len, "[no call, port %5u]: ",
 						log_info.u.stream_fd->socket.local.port);
@@ -66,6 +75,12 @@ static void ilog_prefix_default(char *prefix, size_t prefix_len) {
 					STR_FMT_M(&log_info.u.ice_agent->media->monologue->tag),
 					log_info.u.ice_agent->media->index);
 			break;
+		case LOG_INFO_MEDIA:
+			snprintf(prefix, prefix_len, "[" STR_FORMAT_M "/" STR_FORMAT_M "/%u]: ",
+					STR_FMT_M(&log_info.u.call->callid),
+					STR_FMT_M(&log_info.v.media->monologue->tag),
+					log_info.v.media->index);
+			break;
 	}
 }
 
@@ -79,10 +94,18 @@ static void ilog_prefix_parsable(char *prefix, size_t prefix_len) {
 					STR_FMT(&log_info.u.call->callid));
 			break;
 		case LOG_INFO_STREAM_FD:
-			if (log_info.u.stream_fd->call)
-				snprintf(prefix, prefix_len, "[ID=\""STR_FORMAT"\" port=\"%5u\"]: ",
-						STR_FMT(&log_info.u.stream_fd->call->callid),
-						log_info.u.stream_fd->socket.local.port);
+			if (log_info.u.stream_fd->call) {
+				if (log_info.u.stream_fd->stream)
+					snprintf(prefix, prefix_len, "[ID=\""STR_FORMAT"\" tag=\""STR_FORMAT"\" index=\"%u\" port=\"%5u\"]: ",
+							STR_FMT(&log_info.u.stream_fd->call->callid),
+							STR_FMT(&log_info.u.stream_fd->stream->media->monologue->tag),
+							log_info.u.stream_fd->stream->media->index,
+							log_info.u.stream_fd->socket.local.port);
+				else
+					snprintf(prefix, prefix_len, "[ID=\""STR_FORMAT"\" port=\"%5u\"]: ",
+							STR_FMT(&log_info.u.stream_fd->call->callid),
+							log_info.u.stream_fd->socket.local.port);
+			}
 			break;
 		case LOG_INFO_STR:
 			snprintf(prefix, prefix_len, "[ID=\""STR_FORMAT"\"]: ",
@@ -97,6 +120,12 @@ static void ilog_prefix_parsable(char *prefix, size_t prefix_len) {
 					STR_FMT(&log_info.u.ice_agent->media->monologue->tag),
 					log_info.u.ice_agent->media->index);
 			break;
+		case LOG_INFO_MEDIA:
+			snprintf(prefix, prefix_len, "[ID=\""STR_FORMAT"\" tag=\""STR_FORMAT"\" index=\"%u\"]: ",
+					STR_FMT(&log_info.u.call->callid),
+					STR_FMT(&log_info.v.media->monologue->tag),
+					log_info.v.media->index);
+			break;
 	}
 }
 
@@ -141,6 +170,7 @@ int get_local_log_level(unsigned int subsystem_idx) {
 
 	switch (log_info.e) {
 		case LOG_INFO_CALL:
+		case LOG_INFO_MEDIA:
 			call = log_info.u.call;
 			break;
 		case LOG_INFO_STREAM_FD:
