@@ -39,6 +39,14 @@ static int append_meta_chunk(struct recording *recording, const char *buf, unsig
 		const char *label_fmt, ...)
 	__attribute__((format(printf,4,5)));
 
+// all methods
+static int create_spool_dir_all(const char *spoolpath);
+static void init_all(struct call *call);
+static void sdp_after_all(struct recording *recording, GString *str, struct call_monologue *ml,
+		enum call_opmode opmode);
+static void dump_packet_all(struct media_packet *mp, const str *s);
+static void finish_all(struct call *call);
+
 // pcap methods
 static int rec_pcap_create_spool_dir(const char *dirpath);
 static void rec_pcap_init(struct call *);
@@ -52,7 +60,7 @@ static void proc_init(struct call *);
 static void sdp_before_proc(struct recording *, const str *, struct call_monologue *, enum call_opmode);
 static void sdp_after_proc(struct recording *, GString *str, struct call_monologue *, enum call_opmode opmode);
 static void meta_chunk_proc(struct recording *, const char *, const str *);
-static void update_flags_proc(struct call *call);
+static void update_flags_proc(struct call *call, bool streams);
 static void finish_proc(struct call *);
 static void dump_packet_proc(struct media_packet *mp, const str *s);
 static void init_stream_proc(struct packet_stream *);
@@ -68,15 +76,23 @@ static void rec_pcap_eth_header(unsigned char *, struct packet_stream *);
 #define append_meta_chunk_null(r,f...) append_meta_chunk(r, "", 0, f)
 
 
-static const struct recording_method methods[] = {
+const struct recording_method methods[] = {
 	{
 		.name = "pcap",
 		.kernel_support = 0,
 		.create_spool_dir = rec_pcap_create_spool_dir,
 		.init_struct = rec_pcap_init,
+		.sdp_before = NULL,
 		.sdp_after = sdp_after_pcap,
+		.meta_chunk = NULL,
+		.update_flags = NULL,
 		.dump_packet = dump_packet_pcap,
 		.finish = finish_pcap,
+		.init_stream_struct = NULL,
+		.setup_stream = NULL,
+		.setup_media = NULL,
+		.setup_monologue = NULL,
+		.stream_kernel_info = NULL,
 		.response = response_pcap,
 	},
 	{
@@ -95,6 +111,25 @@ static const struct recording_method methods[] = {
 		.setup_media = setup_media_proc,
 		.setup_monologue = setup_monologue_proc,
 		.stream_kernel_info = kernel_info_proc,
+		.response = NULL,
+	},
+	{
+		.name = "all",
+		.kernel_support = 0,
+		.create_spool_dir = create_spool_dir_all,
+		.init_struct = init_all,
+		.sdp_before = sdp_before_proc,
+		.sdp_after = sdp_after_all,
+		.meta_chunk = meta_chunk_proc,
+		.update_flags = update_flags_proc,
+		.dump_packet = dump_packet_all,
+		.finish = finish_all,
+		.init_stream_struct = init_stream_proc,
+		.setup_stream = setup_stream_proc,
+		.setup_media = setup_media_proc,
+		.setup_monologue = setup_monologue_proc,
+		.stream_kernel_info = kernel_info_proc,
+		.response = response_pcap,
 	},
 };
 
@@ -235,8 +270,8 @@ static int rec_pcap_create_spool_dir(const char *spoolpath) {
 }
 
 // lock must be held
-static void update_metadata(struct call *call, str *metadata) {
-	if (!metadata || !metadata->s)
+void update_metadata_call(struct call *call, str *metadata) {
+	if (!metadata || !metadata->s || !call)
 		return;
 
 	if (str_cmp_str(metadata, &call->metadata)) {
@@ -246,6 +281,22 @@ static void update_metadata(struct call *call, str *metadata) {
 	}
 }
 
+// lock must be held
+void update_metadata_monologue(struct call_monologue *ml, str *metadata) {
+	if (!metadata || !metadata->s || !ml)
+		return;
+
+	struct call *call = ml->call;
+
+	if (str_cmp_str(metadata, &ml->metadata)) {
+		call_str_cpy(call, &ml->metadata, metadata);
+		if (call->recording)
+			append_meta_chunk_str(call->recording, metadata, "METADATA-TAG %u", ml->unique_id);
+	}
+
+	update_metadata_call(call, metadata);
+}
+
 static void update_output_dest(struct call *call, str *output_dest) {
 	if (!output_dest || !output_dest->s || !call->recording)
 		return;
@@ -253,28 +304,28 @@ static void update_output_dest(struct call *call, str *output_dest) {
 }
 
 // lock must be held
-static void update_flags_proc(struct call *call) {
+static void update_flags_proc(struct call *call, bool streams) {
 	append_meta_chunk_null(call->recording, "RECORDING %u", call->recording_on ? 1 : 0);
 	append_meta_chunk_null(call->recording, "FORWARDING %u", call->rec_forwarding ? 1 : 0);
+	if (!streams)
+		return;
 	for (GList *l = call->streams.head; l; l = l->next) {
 		struct packet_stream *ps = l->data;
 		append_meta_chunk_null(call->recording, "STREAM %u FORWARDING %u",
 				ps->unique_id, ps->media->monologue->rec_forwarding ? 1 : 0);
 	}
 }
-static void recording_update_flags(struct call *call) {
-	_rm(update_flags, call);
+static void recording_update_flags(struct call *call, bool streams) {
+	_rm(update_flags, call, streams);
 }
 
 // lock must be held
-void recording_start(struct call *call, const char *prefix, str *metadata, str *output_dest) {
-	update_metadata(call, metadata);
-
+void recording_start(struct call *call, const char *prefix, str *output_dest) {
 	update_output_dest(call, output_dest);
 
 	if (call->recording) {
 		// already active
-		recording_update_flags(call);
+		recording_update_flags(call, true);
 		return;
 	}
 
@@ -299,6 +350,11 @@ void recording_start(struct call *call, const char *prefix, str *metadata, str *
 
 	_rm(init_struct, call);
 
+	// update main call flags (global recording/forwarding on/off) to prevent recording
+	// features from being started when the stream info (through setup_stream) is
+	// propagated if recording is actually off
+	recording_update_flags(call, false);
+
 	// if recording has been turned on after initial call setup, we must walk
 	// through all related objects and initialize the recording stuff. if this
 	// function is called right at the start of the call, all of the following
@@ -319,25 +375,22 @@ void recording_start(struct call *call, const char *prefix, str *metadata, str *
 		__reset_sink_handlers(ps);
 	}
 
-	recording_update_flags(call);
+	recording_update_flags(call, true);
 }
-void recording_stop(struct call *call, str *metadata) {
+void recording_stop(struct call *call) {
 	if (!call->recording)
 		return;
 
-	if (metadata)
-		update_metadata(call, metadata);
-
 	// check if all recording options are disabled
 	if (call->recording_on || call->rec_forwarding) {
-		recording_update_flags(call);
+		recording_update_flags(call, true);
 		return;
 	}
 
 	for (GList *l = call->monologues.head; l; l = l->next) {
 		struct call_monologue *ml = l->data;
 		if (ml->rec_forwarding) {
-			recording_update_flags(call);
+			recording_update_flags(call, true);
 			return;
 		}
 	}
@@ -345,6 +398,10 @@ void recording_stop(struct call *call, str *metadata) {
 	ilog(LOG_NOTICE, "Turning off call recording.");
 	recording_finish(call);
 }
+void recording_pause(struct call *call) {
+	ilog(LOG_NOTICE, "Pausing call recording.");
+	recording_update_flags(call, true);
+}
 
 /**
  *
@@ -356,19 +413,17 @@ void recording_stop(struct call *call, str *metadata) {
  *
  * Returns a boolean for whether or not the call is being recorded.
  */
-void detect_setup_recording(struct call *call, const str *recordcall, str *metadata) {
-	update_metadata(call, metadata);
-
+void detect_setup_recording(struct call *call, const str *recordcall) {
 	if (!recordcall || !recordcall->s)
 		return;
 
 	if (!str_cmp(recordcall, "yes") || !str_cmp(recordcall, "on")) {
 		call->recording_on = 1;
-		recording_start(call, NULL, NULL, NULL);
+		recording_start(call, NULL, NULL);
 	}
 	else if (!str_cmp(recordcall, "no") || !str_cmp(recordcall, "off")) {
 		call->recording_on = 0;
-		recording_stop(call, NULL);
+		recording_stop(call);
 	}
 	else
 		ilog(LOG_INFO, "\"record-call\" flag "STR_FORMAT" is invalid flag.", STR_FMT(recordcall));
@@ -409,14 +464,14 @@ static char *meta_setup_file(struct recording *recording) {
 	}
 
 	char *meta_filepath = file_path_str(recording->meta_prefix, "/tmp/rtpengine-meta-", ".tmp");
-	recording->meta_filepath = meta_filepath;
+	recording->meta_filepath_pcap = meta_filepath;
 	FILE *mfp = fopen(meta_filepath, "w");
 	// coverity[check_return : FALSE]
 	chmod(meta_filepath, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
 	if (mfp == NULL) {
 		ilog(LOG_ERROR, "Could not open metadata file: %s%s%s", FMT_M(meta_filepath));
 		free(meta_filepath);
-		recording->meta_filepath = NULL;
+		recording->meta_filepath_pcap = NULL;
 		return NULL;
 	}
 	recording->u.pcap.meta_fp = mfp;
@@ -497,10 +552,10 @@ static int rec_pcap_meta_finish_file(struct call *call) {
 	// and move it to the finished file location.
 	// Rename extension to ".txt".
 	int fn_len;
-	char *meta_filename = strrchr(recording->meta_filepath, '/');
+	char *meta_filename = strrchr(recording->meta_filepath_pcap, '/');
 	char *meta_ext = NULL;
 	if (meta_filename == NULL) {
-		meta_filename = recording->meta_filepath;
+		meta_filename = recording->meta_filepath_pcap;
 	}
 	else {
 		meta_filename = meta_filename + 1;
@@ -513,13 +568,13 @@ static int rec_pcap_meta_finish_file(struct call *call) {
 	char new_metapath[prefix_len + fn_len + ext_len + 1];
 	snprintf(new_metapath, prefix_len+fn_len+1, "%s/metadata/%s", spooldir, meta_filename);
 	snprintf(new_metapath + prefix_len+fn_len, ext_len+1, ".txt");
-	return_code = return_code || rename(recording->meta_filepath, new_metapath);
+	return_code = return_code || rename(recording->meta_filepath_pcap, new_metapath);
 	if (return_code != 0) {
 		ilog(LOG_ERROR, "Could not move metadata file \"%s\" to \"%s/metadata/\"",
-				 recording->meta_filepath, spooldir);
+				 recording->meta_filepath_pcap, spooldir);
 	} else {
 		ilog(LOG_INFO, "Moved metadata file \"%s\" to \"%s/metadata\"",
-				 recording->meta_filepath, spooldir);
+				 recording->meta_filepath_pcap, spooldir);
 	}
 
 	mutex_destroy(&recording->u.pcap.recording_lock);
@@ -656,7 +711,8 @@ void recording_finish(struct call *call) {
 
 	free(recording->meta_prefix);
 	free(recording->escaped_callid);
-	free(recording->meta_filepath);
+	free(recording->meta_filepath_pcap);
+	free(recording->meta_filepath_proc);
 
 	g_slice_free1(sizeof(*(recording)), recording);
 	call->recording = NULL;
@@ -671,10 +727,10 @@ void recording_finish(struct call *call) {
 
 static int open_proc_meta_file(struct recording *recording) {
 	int fd;
-	fd = open(recording->meta_filepath, O_WRONLY | O_APPEND | O_CREAT, 0666);
+	fd = open(recording->meta_filepath_proc, O_WRONLY | O_APPEND | O_CREAT, 0666);
 	if (fd == -1) {
 		ilog(LOG_ERR, "Failed to open recording metadata file '%s' for writing: %s",
-				recording->meta_filepath, strerror(errno));
+				recording->meta_filepath_proc, strerror(errno));
 		return -1;
 	}
 	return fd;
@@ -740,8 +796,8 @@ static void proc_init(struct call *call) {
 	}
 	ilog(LOG_DEBUG, "kernel call idx is %u", recording->u.proc.call_idx);
 
-	recording->meta_filepath = file_path_str(recording->meta_prefix, "/", ".meta");
-	unlink(recording->meta_filepath); // start fresh XXX good idea?
+	recording->meta_filepath_proc = file_path_str(recording->meta_prefix, "/", ".meta");
+	unlink(recording->meta_filepath_proc); // start fresh XXX good idea?
 
 	append_meta_chunk_str(recording, &call->callid, "CALL-ID");
 	append_meta_chunk_s(recording, recording->meta_prefix, "PARENT");
@@ -775,7 +831,7 @@ static void finish_proc(struct call *call) {
 		struct packet_stream *ps = l->data;
 		ps->recording.u.proc.stream_idx = UNINIT_IDX;
 	}
-	unlink(recording->meta_filepath);
+	unlink(recording->meta_filepath_proc);
 }
 
 static void init_stream_proc(struct packet_stream *stream) {
@@ -825,6 +881,8 @@ static void setup_monologue_proc(struct call_monologue *ml) {
 	append_meta_chunk_str(recording, &ml->tag, "TAG %u", ml->unique_id);
 	if (ml->label.len)
 		append_meta_chunk_str(recording, &ml->label, "LABEL %u", ml->unique_id);
+	if (ml->metadata.len)
+		append_meta_chunk_str(recording, &ml->metadata, "METADATA-TAG %u", ml->unique_id);
 }
 
 static void setup_media_proc(struct call_media *media) {
@@ -886,3 +944,38 @@ static void kernel_info_proc(struct packet_stream *stream, struct rtpengine_targ
 static void meta_chunk_proc(struct recording *recording, const char *label, const str *data) {
 	append_meta_chunk_str(recording, data, "%s", label);
 }
+
+static int create_spool_dir_all(const char *spoolpath) {
+	int ret1, ret2;
+
+	ret1 = rec_pcap_create_spool_dir(spoolpath);
+	ret2 = check_main_spool_dir(spoolpath);
+
+	if (ret1 == FALSE || ret2 == FALSE) {
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+static void init_all(struct call *call) {
+	rec_pcap_init(call);
+	proc_init(call);
+}
+
+static void sdp_after_all(struct recording *recording, GString *str, struct call_monologue *ml,
+		enum call_opmode opmode)
+{
+	sdp_after_pcap(recording, str, ml, opmode);
+	sdp_after_proc(recording, str, ml, opmode);
+}
+
+static void dump_packet_all(struct media_packet *mp, const str *s) {
+	dump_packet_pcap(mp, s);
+	dump_packet_proc(mp, s);
+}
+
+static void finish_all(struct call *call) {
+	finish_pcap(call);
+	finish_proc(call);
+}
