@@ -256,13 +256,13 @@ void __t38_gateway_free(void *p) {
 }
 
 // call is locked in R and mp is locked
-static void t38_pcm_player(struct media_player *mp) {
+static bool t38_pcm_player(struct media_player *mp) {
 	if (!mp || !mp->media)
-		return;
+		return true;
 
 	struct t38_gateway *tg = mp->media->t38_gateway;
 	if (!tg)
-		return;
+		return true;
 
 	if (tg->pcm_media && tg->pcm_media->streams.head
 			&& ((struct packet_stream *) tg->pcm_media->streams.head->data)->selected_sfd)
@@ -279,7 +279,7 @@ static void t38_pcm_player(struct media_player *mp) {
 		timeval_add_usec(&mp->next_run, 10000);
 		timerthread_obj_schedule_abs(&mp->tt_obj, &mp->next_run);
 		mutex_unlock(&tg->lock);
-		return;
+		return false;
 	}
 
 	ilog(LOG_DEBUG, "Generated %i T.38 PCM samples", num);
@@ -295,6 +295,8 @@ static void t38_pcm_player(struct media_player *mp) {
 	// this reschedules our player as well
 	media_player_add_packet(pcm_player, (char *) smp, num * 2, num * 1000000 / 8000, pts);
 	media_player_put(&pcm_player);
+
+	return false;
 }
 
 
@@ -375,6 +377,7 @@ int t38_gateway_pair(struct call_media *t38_media, struct call_media *pcm_media,
 	tg->pcm_pt.encoding_with_params = tg->pcm_pt.encoding;
 	tg->pcm_pt.clock_rate = 8000;
 	tg->pcm_pt.channels = 1;
+	tg->pcm_pt.ptime = 20;
 
 	err = "Failed to init PCM codec";
 	ensure_codec_def(&tg->pcm_pt, pcm_media);
@@ -458,7 +461,7 @@ void t38_gateway_start(struct t38_gateway *tg) {
 
 	// set up our player first
 	media_player_set_media(tg->pcm_player, tg->pcm_media);
-	if (media_player_setup(tg->pcm_player, &tg->pcm_pt))
+	if (media_player_setup(tg->pcm_player, &tg->pcm_pt, NULL))
 		return;
 
 	// now start our player if we can or should
@@ -467,7 +470,7 @@ void t38_gateway_start(struct t38_gateway *tg) {
 		return;
 
 	// only start our player only if we can send both ways
-	if (!tg->pcm_media->codecs_prefs_send.length)
+	if (!tg->pcm_media->codecs.codec_prefs.length)
 		return;
 	if (!tg->pcm_media->streams.length)
 		return;
@@ -568,11 +571,11 @@ int t38_gateway_input_udptl(struct t38_gateway *tg, const str *buf) {
 		return 0;
 
 	if (buf->len < 4) {
-		ilog(LOG_INFO | LOG_FLAG_LIMIT, "Ignoring short UDPTL packet (%i bytes)", buf->len);
+		ilog(LOG_INFO | LOG_FLAG_LIMIT, "Ignoring short UDPTL packet (%zu bytes)", buf->len);
 		return 0;
 	}
 
-	ilog(LOG_DEBUG, "Processing %i UDPTL bytes", buf->len);
+	ilog(LOG_DEBUG, "Processing %zu UDPTL bytes", buf->len);
 
 	str s = *buf;
 	str piece;
@@ -587,7 +590,7 @@ int t38_gateway_input_udptl(struct t38_gateway *tg, const str *buf) {
 	if (__get_udptl(&piece, &s))
 		goto err;
 
-	ilog(LOG_DEBUG, "Received primary IFP packet, len %i, seq %i", piece.len, seq);
+	ilog(LOG_DEBUG, "Received primary IFP packet, len %zu, seq %i", piece.len, seq);
 	str primary = piece;
 	up = __make_udptl_packet(&primary, seq);
 
@@ -636,7 +639,7 @@ int t38_gateway_input_udptl(struct t38_gateway *tg, const str *buf) {
 			// ignore zero-length packets
 			if (!piece.len)
 				continue;
-			ilog(LOG_DEBUG, "Received secondary IFP packet, len %i, seq %i", piece.len,
+			ilog(LOG_DEBUG, "Received secondary IFP packet, len %zu, seq %i", piece.len,
 					seq - 1 - i);
 			up = __make_udptl_packet(&piece, seq - 1 - i);
 			packet_sequencer_insert(&tg->sequencer, &up->p);
@@ -744,7 +747,7 @@ seq_ok:;
 		if (!up)
 			break;
 
-		ilog(LOG_DEBUG, "Processing %i IFP bytes, seq %i", up->s->len, up->p.seq);
+		ilog(LOG_DEBUG, "Processing %zu IFP bytes, seq %i", up->s->len, up->p.seq);
 
 		t38_core_rx_ifp_packet(t38, (uint8_t *) up->s->s, up->s->len, up->p.seq);
 
