@@ -19,6 +19,9 @@ struct websocket_output {
 	GString *str;
 	size_t str_done;
 	enum lws_write_protocol protocol;
+	int http_status;
+	const char *content_type;
+	ssize_t content_length;
 };
 
 struct websocket_conn {
@@ -51,6 +54,8 @@ struct websocket_ng_buf {
 static GQueue websocket_vhost_configs;
 static struct lws_context *websocket_context;
 static GThreadPool *websocket_threads;
+static mutex_t websocket_callback_lock = MUTEX_STATIC_INIT;
+static mutex_t websocket_service_lock = MUTEX_STATIC_INIT;
 
 
 static struct websocket_message *websocket_message_new(struct websocket_conn *wc) {
@@ -103,27 +108,27 @@ static void __websocket_queue_raw(struct websocket_conn *wc, const char *msg, si
 
 // appends to output buffer without triggering a response
 void websocket_queue_raw(struct websocket_conn *wc, const char *msg, size_t len) {
-	mutex_lock(&wc->lock);
+	LOCK(&wc->lock);
 	__websocket_queue_raw(wc, msg, len);
-	mutex_unlock(&wc->lock);
 }
 
 
 // num bytes in output buffer
 size_t websocket_queue_len(struct websocket_conn *wc) {
-	mutex_lock(&wc->lock);
+	LOCK(&wc->lock);
+
 	size_t ret = 0;
 	for (GList *l = wc->output_q.head; l; l = l->next) {
 		struct websocket_output *wo = l->data;
 		ret += (wo->str->len - LWS_PRE);
 	}
-	mutex_unlock(&wc->lock);
+
 	return ret;
 }
 
 
 // adds data to output buffer (can be null) and optionally triggers specified response
-int websocket_write_raw(struct websocket_conn *wc, const char *msg, size_t len,
+void websocket_write_raw(struct websocket_conn *wc, const char *msg, size_t len,
 		enum lws_write_protocol protocol, bool done)
 {
 	mutex_lock(&wc->lock);
@@ -132,36 +137,63 @@ int websocket_write_raw(struct websocket_conn *wc, const char *msg, size_t len,
 	wo->protocol = protocol;
 	g_queue_push_tail(&wc->output_q, websocket_output_new());
 
-	if (done) {
-		lws_callback_on_writable(wc->wsi);
-		lws_cancel_service(websocket_context);
-	}
-
 	mutex_unlock(&wc->lock);
 
-	return 0;
+	if (done) {
+		// Sadly lws_callback_on_writable() doesn't do any internal
+		// locking, therefore we must protect it against a concurrently
+		// running lws_service(), as well as against other threads
+		// invoking lws_callback_on_writable().
+		//
+		// Acquire the callback lock first, which is normally unlocked,
+		// then wake up the service thread and try to break out of
+		// lws_service().  The service thread holds the service lock
+		// while lws_service() is executing and releases it as soon as
+		// lws_service() is done. We therefore try to acquire the
+		// service lock here, which blocks us until lws_service() is
+		// actually done. At this point the service thread will try to
+		// acquire the callback lock, which is still held by us here,
+		// and so the service thread will block until we are done
+		// calling lws_callback_on_writable(). Finally we release both
+		// locks, which allows the service thread to resume
+		// lws_service().
+		//
+		// The suggested approach of using
+		// LWS_CALLBACK_EVENT_WAIT_CANCELLED together with a queue and
+		// then calling lws_callback_on_writable() from the service
+		// thread is not usable as libwebsockets 2.0 doesn't support
+		// LWS_CALLBACK_EVENT_WAIT_CANCELLED.
+
+		mutex_lock(&websocket_callback_lock);
+		lws_cancel_service(websocket_context);
+
+		mutex_lock(&websocket_service_lock);
+		lws_callback_on_writable(wc->wsi);
+
+		mutex_unlock(&websocket_service_lock);
+		mutex_unlock(&websocket_callback_lock);
+	}
 }
 
 
 // adds data to output buffer (can be null) and triggers specified response: http or binary websocket
-int websocket_write_http_len(struct websocket_conn *wc, const char *msg, size_t len, bool done) {
-	return websocket_write_raw(wc, msg, len, LWS_WRITE_HTTP, done);
+void websocket_write_http_len(struct websocket_conn *wc, const char *msg, size_t len, bool done) {
+	websocket_write_raw(wc, msg, len, LWS_WRITE_HTTP, done);
 }
-int websocket_write_http(struct websocket_conn *wc, const char *msg, bool done) {
-	return websocket_write_http_len(wc, msg, msg ? strlen(msg) : 0, done);
+void websocket_write_http(struct websocket_conn *wc, const char *msg, bool done) {
+	websocket_write_http_len(wc, msg, msg ? strlen(msg) : 0, done);
 }
-int websocket_write_text(struct websocket_conn *wc, const char *msg, bool done) {
-	return websocket_write_raw(wc, msg, strlen(msg), LWS_WRITE_TEXT, done);
+void websocket_write_text(struct websocket_conn *wc, const char *msg, bool done) {
+	websocket_write_raw(wc, msg, strlen(msg), LWS_WRITE_TEXT, done);
 }
-int websocket_write_binary(struct websocket_conn *wc, const char *msg, size_t len, bool done) {
-	return websocket_write_raw(wc, msg, len, LWS_WRITE_BINARY, done);
+void websocket_write_binary(struct websocket_conn *wc, const char *msg, size_t len, bool done) {
+	websocket_write_raw(wc, msg, len, LWS_WRITE_BINARY, done);
 }
 
 
 void websocket_write_next(struct websocket_conn *wc) {
-	mutex_lock(&wc->lock);
+	LOCK(&wc->lock);
 	g_queue_push_tail(&wc->output_q, websocket_output_new());
-	mutex_unlock(&wc->lock);
 }
 
 
@@ -176,7 +208,7 @@ static const char *websocket_echo_process(struct websocket_message *wm) {
 static void websocket_message_push(struct websocket_conn *wc, websocket_message_func_t func) {
 	ilogs(http, LOG_DEBUG, "Adding HTTP/WS message to processing queue");
 
-	mutex_lock(&wc->lock);
+	LOCK(&wc->lock);
 
 	struct websocket_message *wm = wc->wm;
 	assert(wm != NULL);
@@ -187,8 +219,6 @@ static void websocket_message_push(struct websocket_conn *wc, websocket_message_
 	g_thread_pool_push(websocket_threads, wc, NULL);
 
 	wc->wm = websocket_message_new(wc);
-
-	mutex_unlock(&wc->lock);
 }
 
 
@@ -219,54 +249,8 @@ static void websocket_process(void *p, void *up) {
 }
 
 
-static int websocket_dequeue(struct websocket_conn *wc) {
-	if (!wc)
-		return 0;
-
-	int is_http = 0;
-
-	mutex_lock(&wc->lock);
-	struct websocket_output *wo;
-	struct lws *wsi = wc->wsi;
-	while ((wo = g_queue_pop_head(&wc->output_q))) {
-		// used buffer slot?
-		if (wo->str) {
-			// allocate post-buffer
-			g_string_set_size(wo->str, wo->str->len + LWS_SEND_BUFFER_POST_PADDING);
-			size_t to_send = wo->str->len - wo->str_done - LWS_SEND_BUFFER_POST_PADDING;
-			if (to_send) {
-				if (to_send > 2000)
-					ilogs(http, LOG_DEBUG, "Writing %lu bytes to LWS", (unsigned long) to_send);
-				else
-					ilogs(http, LOG_DEBUG, "Writing back to LWS: '%.*s'",
-							(int) to_send, wo->str->str + wo->str_done);
-				size_t ret = lws_write(wsi, (unsigned char *) wo->str->str + wo->str_done,
-						to_send, wo->protocol);
-				if (ret != to_send)
-					ilogs(http, LOG_ERR, "Invalid LWS write: %lu != %lu",
-							(unsigned long) ret,
-							(unsigned long) to_send);
-				wo->str_done += ret;
-
-				if (wo->protocol == LWS_WRITE_HTTP)
-					is_http = 1;
-			}
-		}
-		websocket_output_free(wo);
-	}
-	g_queue_push_tail(&wc->output_q, websocket_output_new());
-
-	mutex_unlock(&wc->lock);
-
-	int ret = 0;
-	if (is_http)
-		ret = lws_http_transaction_completed(wsi); // may destroy `wc`
-
-	return ret;
-}
-
-static const char *websocket_do_http_response(struct websocket_conn *wc, int status, const char *content_type,
-		ssize_t content_length)
+static const char *__websocket_write_http_response(struct websocket_conn *wc, int status,
+		const char *content_type, ssize_t content_length)
 {
 	uint8_t buf[LWS_PRE + 2048], *start = &buf[LWS_PRE], *p = start,
 		*end = &buf[sizeof(buf) - LWS_PRE - 1];
@@ -281,6 +265,24 @@ static const char *websocket_do_http_response(struct websocket_conn *wc, int sta
 	if (content_length >= 0)
 		if (lws_add_http_header_content_length(wc->wsi, content_length, &p, end))
 			return "Failed to add HTTP headers to response";
+
+	if (lws_add_http_header_by_name(wc->wsi,
+				(unsigned char *) "Access-Control-Allow-Origin:",
+				(unsigned char *) "*", 1, &p, end))
+		return "Failed to add CORS header";
+	if (lws_add_http_header_by_name(wc->wsi,
+				(unsigned char *) "Access-Control-Max-Age:",
+				(unsigned char *) "86400", 5, &p, end))
+		return "Failed to add CORS header";
+	if (lws_add_http_header_by_name(wc->wsi,
+				(unsigned char *) "Access-Control-Allow-Methods:",
+				(unsigned char *) "GET, POST", 9, &p, end))
+		return "Failed to add CORS header";
+	if (lws_add_http_header_by_name(wc->wsi,
+				(unsigned char *) "Access-Control-Allow-Headers:",
+				(unsigned char *) "Content-Type", 12, &p, end))
+		return "Failed to add CORS header";
+
 	if (lws_finalize_http_header(wc->wsi, &p, end))
 		return "Failed to write HTTP headers";
 
@@ -290,36 +292,95 @@ static const char *websocket_do_http_response(struct websocket_conn *wc, int sta
 
 	return NULL;
 }
-int websocket_http_response(struct websocket_conn *wc, int status, const char *content_type,
+static int websocket_dequeue(struct websocket_conn *wc) {
+	if (!wc)
+		return 0;
+
+	int is_http = 0;
+
+	mutex_lock(&wc->lock);
+	struct websocket_output *wo;
+	struct lws *wsi = wc->wsi;
+	while ((wo = g_queue_pop_head(&wc->output_q))) {
+		// used buffer slot?
+		if (!wo->str)
+			goto next;
+
+		if (wo->http_status) {
+			const char *err = __websocket_write_http_response(wc, wo->http_status,
+					wo->content_type, wo->content_length);
+			if (err) {
+				ilogs(http, LOG_ERR, "Failed to write HTTP response headers: %s", err);
+				goto next;
+			}
+		}
+
+		// allocate post-buffer
+		g_string_set_size(wo->str, wo->str->len + LWS_SEND_BUFFER_POST_PADDING);
+		size_t to_send = wo->str->len - wo->str_done - LWS_SEND_BUFFER_POST_PADDING;
+		if (to_send) {
+			if (to_send > 2000)
+				ilogs(http, LOG_DEBUG, "Writing %lu bytes to LWS", (unsigned long) to_send);
+			else
+				ilogs(http, LOG_DEBUG, "Writing back to LWS: '%.*s'",
+						(int) to_send, wo->str->str + wo->str_done);
+			size_t ret = lws_write(wsi, (unsigned char *) wo->str->str + wo->str_done,
+					to_send, wo->protocol);
+			if (ret != to_send)
+				ilogs(http, LOG_ERR, "Invalid LWS write: %lu != %lu",
+						(unsigned long) ret,
+						(unsigned long) to_send);
+			wo->str_done += ret;
+
+			if (wo->protocol == LWS_WRITE_HTTP)
+				is_http = 1;
+		}
+
+next:
+		websocket_output_free(wo);
+	}
+	g_queue_push_tail(&wc->output_q, websocket_output_new());
+
+	mutex_unlock(&wc->lock);
+
+	int ret = 0;
+	if (is_http)
+		if (lws_http_transaction_completed(wsi) == 1) // may destroy `wc`
+			ret = -1;
+
+	return ret;
+}
+
+void websocket_http_response(struct websocket_conn *wc, int status, const char *content_type,
 		ssize_t content_length)
 {
-	const char *err = websocket_do_http_response(wc, status, content_type, content_length);
-	if (!err)
-		return 0;
-	ilogs(http, LOG_ERR, "Failed to write HTTP response headers: %s", err);
-	return -1;
+	LOCK(&wc->lock);
+
+	struct websocket_output *wo = g_queue_peek_tail(&wc->output_q);
+
+	wo->http_status = status;
+	wo->content_type = content_type;
+	wo->content_length = content_length;
 }
-const char *websocket_http_complete(struct websocket_conn *wc, int status, const char *content_type,
+void websocket_http_complete(struct websocket_conn *wc, int status, const char *content_type,
 		ssize_t content_length, const char *content)
 {
-	if (websocket_http_response(wc, status, content_type, content_length))
-		return "Failed to write response HTTP headers";
-	if (websocket_write_http(wc, content, true))
-		return "Failed to write pong response";
-	return NULL;
+	websocket_http_response(wc, status, content_type, content_length);
+	websocket_write_http(wc, content, true);
 }
 
 
 static const char *websocket_http_ping(struct websocket_message *wm) {
 	ilogs(http, LOG_DEBUG, "Respoding to GET /ping");
-	return websocket_http_complete(wm->wc, 200, "text/plain", 5, "pong\n");
+	websocket_http_complete(wm->wc, 200, "text/plain", 5, "pong\n");
+	return NULL;
 }
 
 
 static const char *websocket_http_metrics(struct websocket_message *wm) {
 	ilogs(http, LOG_DEBUG, "Respoding to GET /metrics");
 
-	AUTO_CLEANUP_INIT(GQueue *metrics, statistics_free_metrics, statistics_gather_metrics());
+	AUTO_CLEANUP_INIT(GQueue *metrics, statistics_free_metrics, statistics_gather_metrics(NULL));
 	AUTO_CLEANUP_INIT(GString *outp, __g_string_free, g_string_new(""));
 	AUTO_CLEANUP_INIT(GHashTable *metric_types, __g_hash_table_destroy,
 			g_hash_table_new(g_str_hash, g_str_equal));
@@ -349,7 +410,8 @@ static const char *websocket_http_metrics(struct websocket_message *wm) {
 		g_string_append_printf(outp, " %s\n", m->value_short);
 	}
 
-	return websocket_http_complete(wm->wc, 200, "text/plain", outp->len, outp->str);
+	websocket_http_complete(wm->wc, 200, "text/plain", outp->len, outp->str);
+	return NULL;
 }
 
 
@@ -381,7 +443,8 @@ static const char *websocket_http_cli(struct websocket_message *wm) {
 
 	size_t len = websocket_queue_len(wm->wc);
 
-	return websocket_http_complete(wm->wc, 200, "text/plain", len, NULL);
+	websocket_http_complete(wm->wc, 200, "text/plain", len, NULL);
+	return NULL;
 }
 
 
@@ -411,8 +474,7 @@ static void websocket_ng_send_ws(str *cookie, str *body, const endpoint_t *sin,
 }
 static void websocket_ng_send_http(str *cookie, str *body, const endpoint_t *sin, void *p1) {
 	struct websocket_conn *wc = p1;
-	if (websocket_http_response(wc, 200, "application/x-rtpengine-ng", cookie->len + 1 + body->len))
-		ilogs(http, LOG_WARN, "Failed to write HTTP headers");
+	websocket_http_response(wc, 200, "application/x-rtpengine-ng", cookie->len + 1 + body->len);
 	websocket_queue_raw(wc, cookie->s, cookie->len);
 	websocket_queue_raw(wc, " ", 1);
 	websocket_queue_raw(wc, body->s, body->len);
@@ -472,7 +534,7 @@ static int websocket_http_get(struct websocket_conn *wc) {
 	const char *uri = wm->uri;
 	websocket_message_func_t handler = NULL;
 
-	ilogs(http, LOG_INFO, "HTTP GET from %s: '%s'", endpoint_print_buf(&wc->endpoint), wm->uri);
+	ilogs(http, LOG_INFO, "HTTP GET from %s: '%s'", endpoint_print_buf(&wc->endpoint), uri);
 	wm->method = M_GET;
 
 	if (!strcmp(uri, "/ping"))
@@ -481,6 +543,8 @@ static int websocket_http_get(struct websocket_conn *wc) {
 		handler = websocket_http_cli;
 	else if (!strcmp(uri, "/metrics"))
 		handler = websocket_http_metrics;
+	else if (!strncmp(uri, "/admin/", 7))
+		handler = websocket_janus_get;
 
 	if (!handler) {
 		ilogs(http, LOG_WARN, "Unhandled HTTP GET URI: '%s'", uri);
@@ -528,6 +592,25 @@ static int websocket_http_post(struct websocket_conn *wc) {
 }
 
 
+static const char *websocket_http_options_generic(struct websocket_message *wm) {
+	ilogs(http, LOG_DEBUG, "Respoding to OPTIONS");
+	websocket_http_complete(wm->wc, 200, NULL, 0, NULL);
+	return NULL;
+}
+
+
+static int websocket_http_options(struct websocket_conn *wc) {
+	struct websocket_message *wm = wc->wm;
+
+	ilogs(http, LOG_INFO, "HTTP OPTIONS from %s: '%s'", endpoint_print_buf(&wc->endpoint), wm->uri);
+	wm->method = M_OPTIONS;
+
+	websocket_message_push(wc, websocket_http_options_generic);
+
+	return 0;
+}
+
+
 static int websocket_http_body(struct websocket_conn *wc, const char *body, size_t len) {
 	struct websocket_message *wm = wc->wm;
 	const char *uri = wm->uri;
@@ -550,6 +633,10 @@ static int websocket_http_body(struct websocket_conn *wc, const char *body, size
 		handler = websocket_http_ng;
 	else if (!strcmp(uri, "/admin") && wm->method == M_POST && wm->content_type == CT_JSON)
 		handler = websocket_janus_process;
+	else if (!strcmp(uri, "/janus") && wm->method == M_POST && wm->content_type == CT_JSON)
+		handler = websocket_janus_process;
+	else if (!strncmp(uri, "/janus/", 7) && wm->method == M_POST && wm->content_type == CT_JSON)
+		handler = websocket_janus_post;
 
 	if (!handler) {
 		ilogs(http, LOG_WARN, "Unhandled HTTP POST URI: '%s'", wm->uri);
@@ -619,12 +706,6 @@ static int websocket_conn_init(struct lws *wsi, void *p) {
 		return -1;
 
 	memset(wc, 0, sizeof(*wc));
-	wc->wsi = wsi;
-	mutex_init(&wc->lock);
-	cond_init(&wc->cond);
-	g_queue_init(&wc->messages);
-	g_queue_push_tail(&wc->output_q, websocket_output_new());
-	wc->janus_sessions = g_hash_table_new(g_direct_hash, g_direct_equal);
 
 	struct sockaddr_storage sa = {0,};
 	socklen_t sl = sizeof(sa);
@@ -645,22 +726,20 @@ static int websocket_conn_init(struct lws *wsi, void *p) {
 	int fd = lws_get_socket_fd(wsi);
 #endif
 
-	memset(wc, 0, sizeof(*wc));
-
 	if (getpeername(fd, (struct sockaddr *) &sa, &sl)) {
 		ilogs(http, LOG_ERR, "Failed to get remote address of HTTP/WS connection (fd %i): %s",
 				fd, strerror(errno));
 		return -1;
-	} else {
-		endpoint_parse_sockaddr_storage(&wc->endpoint, &sa);
 	}
 
+	endpoint_parse_sockaddr_storage(&wc->endpoint, &sa);
 	wc->wsi = wsi;
 	mutex_init(&wc->lock);
 	cond_init(&wc->cond);
 	g_queue_init(&wc->messages);
 	g_queue_push_tail(&wc->output_q, websocket_output_new());
 	wc->wm = websocket_message_new(wc);
+	wc->janus_sessions = g_hash_table_new(g_direct_hash, g_direct_equal);
 
 	return 0;
 }
@@ -687,6 +766,8 @@ static int websocket_do_http(struct lws *wsi, struct websocket_conn *wc, const c
 		return websocket_http_get(wc);
 	if (lws_hdr_total_length(wsi, WSI_TOKEN_POST_URI))
 		return websocket_http_post(wc);
+	if (lws_hdr_total_length(wsi, WSI_TOKEN_OPTIONS_URI))
+		return websocket_http_options(wc);
 
 	ilogs(http, LOG_INFO, "Ignoring HTTP request to %s with unsupported method", uri);
 	return 0;
@@ -757,6 +838,8 @@ static int websocket_http(struct lws *wsi, enum lws_callback_reasons reason, voi
 			break;
 	}
 
+	release_closed_sockets();
+
 	return 0;
 }
 
@@ -826,6 +909,8 @@ static int websocket_protocol(struct lws *wsi, enum lws_callback_reasons reason,
 			break;
 	}
 
+	release_closed_sockets();
+
 	return 0;
 }
 
@@ -907,12 +992,33 @@ static void websocket_cleanup(void) {
 
 	while (websocket_vhost_configs.length) {
 		struct lws_context_creation_info *vhost = g_queue_pop_head(&websocket_vhost_configs);
-		free((void *) vhost->iface);
-		free(vhost);
+		g_free((void *) vhost->iface);
+		g_slice_free1(sizeof(*vhost), vhost);
 	}
 }
 
 
+static void addr_any_v6_consolidate(endpoint_t eps[2], bool have_lws_ipv6) {
+	// Don't try to double bind on ADDR_ANY. If we find ADDR_ANY, bind to the
+	// v6 port and omit the v4 binding (and let libwebsockets handle the 4/6
+	// translation) unless we find ourselves without v6 support.
+
+	if (!eps[1].port)
+		return;
+	if (!have_lws_ipv6)
+		return;
+	if (eps[0].address.family->af == AF_INET6)
+		return;
+	if (!is_addr_unspecified(&eps[0].address))
+		return;
+
+	// The only case that needs handling: ADDR_ANY requested, v6 support is
+	// available, and v6 binding is given second.
+
+	eps[0] = eps[1];
+	eps[1].port = 0;
+}
+
 int websocket_init(void) {
 	assert(websocket_context == NULL);
 
@@ -920,6 +1026,10 @@ int websocket_init(void) {
 			(!rtpe_config.https_ifs || !*rtpe_config.https_ifs))
 		return 0;
 
+	struct sockaddr_storage sa;
+	int ret = lws_interface_to_sa(1, "::1", (void *) &sa, sizeof(sa));
+	bool have_lws_ipv6 = (ret == 0);
+
 	const char *err = NULL;
 
 	lws_set_log_level(LLL_ERR | LLL_WARN, websocket_log);
@@ -939,24 +1049,43 @@ int websocket_init(void) {
 	for (char **ifp = rtpe_config.http_ifs; ifp && *ifp; ifp++) {
 		char *ifa = *ifp;
 		ilogs(http, LOG_DEBUG, "Starting HTTP/WS '%s'", ifa);
-		endpoint_t ep;
+		endpoint_t eps[2];
 		err = "Failed to parse address/port";
-		if (endpoint_parse_any_getaddrinfo(&ep, ifa))
+		if (endpoint_parse_any_getaddrinfo_alt(&eps[0], &eps[1], ifa))
 			goto err;
+		addr_any_v6_consolidate(eps, have_lws_ipv6);
 
-		struct lws_context_creation_info *vhost = malloc(sizeof(*vhost));
-		g_queue_push_tail(&websocket_vhost_configs, vhost);
+		bool success = false;
+		bool ipv6_fail = false;
+		for (int i = 0; i < G_N_ELEMENTS(eps); i++) {
+			endpoint_t *ep = &eps[i];
+			if (!ep->port)
+				continue;
+			if (ep->address.family->af == AF_INET6 && !have_lws_ipv6) {
+				ipv6_fail = true;
+				continue;
+			}
+			struct lws_context_creation_info *vhost = g_slice_alloc(sizeof(*vhost));
+			g_queue_push_tail(&websocket_vhost_configs, vhost);
 
-		*vhost = (struct lws_context_creation_info) {
-			.port = ep.port,
-			.iface = strdup(sockaddr_print_buf(&ep.address)),
-			.protocols = websocket_protocols,
-		};
-		vhost->vhost_name = vhost->iface;
-		if (ep.address.family->af == AF_INET)
-			vhost->options |= LWS_SERVER_OPTION_DISABLE_IPV6;
-		err = "LWS failed to create vhost";
-		if (!lws_create_vhost(websocket_context, vhost))
+			*vhost = (struct lws_context_creation_info) {
+				.port = ep->port,
+				.iface = g_strdup(sockaddr_print_buf(&ep->address)),
+				.protocols = websocket_protocols,
+			};
+			vhost->vhost_name = vhost->iface;
+			if (ep->address.family->af == AF_INET)
+				vhost->options |= LWS_SERVER_OPTION_DISABLE_IPV6;
+			err = "LWS failed to create vhost";
+			if (!lws_create_vhost(websocket_context, vhost))
+				goto err;
+			success = true;
+		}
+		err = "Failed to create any LWS vhost from given config";
+		if (ipv6_fail)
+			err = "Failed to create any LWS vhost from given config. Hint: LWS IPv6 support is not "
+				"available and config lists at least one IPv6 vhost";
+		if (!success)
 			goto err;
 	}
 
@@ -967,28 +1096,47 @@ int websocket_init(void) {
 
 		char *ifa = *ifp;
 		ilogs(http, LOG_DEBUG, "Starting HTTPS/WSS '%s'", ifa);
-		endpoint_t ep;
+		endpoint_t eps[2];
 		err = "Failed to parse address/port";
-		if (endpoint_parse_any_getaddrinfo(&ep, ifa))
+		if (endpoint_parse_any_getaddrinfo_alt(&eps[0], &eps[1], ifa))
 			goto err;
+		addr_any_v6_consolidate(eps, have_lws_ipv6);
 
-		struct lws_context_creation_info *vhost = malloc(sizeof(*vhost));
-		g_queue_push_tail(&websocket_vhost_configs, vhost);
+		bool success = false;
+		bool ipv6_fail = false;
+		for (int i = 0; i < G_N_ELEMENTS(eps); i++) {
+			endpoint_t *ep = &eps[i];
+			if (!ep->port)
+				continue;
+			if (ep->address.family->af == AF_INET6 && !have_lws_ipv6) {
+				ipv6_fail = true;
+				continue;
+			}
+			struct lws_context_creation_info *vhost = g_slice_alloc(sizeof(*vhost));
+			g_queue_push_tail(&websocket_vhost_configs, vhost);
 
-		*vhost = (struct lws_context_creation_info) {
-			.port = ep.port,
-			.iface = strdup(sockaddr_print_buf(&ep.address)),
-			.protocols = websocket_protocols,
-			.ssl_cert_filepath = rtpe_config.https_cert,
-			.ssl_private_key_filepath = rtpe_config.https_key ? : rtpe_config.https_cert,
-			.options = LWS_SERVER_OPTION_DO_SSL_GLOBAL_INIT,
-			// XXX cipher list, key password
-		};
-		vhost->vhost_name = vhost->iface;
-		if (ep.address.family->af == AF_INET)
-			vhost->options |= LWS_SERVER_OPTION_DISABLE_IPV6;
-		err = "LWS failed to create vhost";
-		if (!lws_create_vhost(websocket_context, vhost))
+			*vhost = (struct lws_context_creation_info) {
+				.port = ep->port,
+				.iface = g_strdup(sockaddr_print_buf(&ep->address)),
+				.protocols = websocket_protocols,
+				.ssl_cert_filepath = rtpe_config.https_cert,
+				.ssl_private_key_filepath = rtpe_config.https_key ? : rtpe_config.https_cert,
+				.options = LWS_SERVER_OPTION_DO_SSL_GLOBAL_INIT,
+				// XXX cipher list, key password
+			};
+			vhost->vhost_name = vhost->iface;
+			if (ep->address.family->af == AF_INET)
+				vhost->options |= LWS_SERVER_OPTION_DISABLE_IPV6;
+			err = "LWS failed to create vhost";
+			if (!lws_create_vhost(websocket_context, vhost))
+				goto err;
+			success = true;
+		}
+		err = "Failed to create any LWS vhost from given config";
+		if (ipv6_fail)
+			err = "Failed to create any LWS vhost from given config. Hint: LWS IPv6 support is not "
+				"available and config lists at least one IPv6 vhost";
+		if (!success)
 			goto err;
 	}
 
@@ -1006,8 +1154,16 @@ err:
 
 static void websocket_loop(void *p) {
 	ilogs(http, LOG_INFO, "Websocket listener thread running");
-	while (!rtpe_shutdown)
+	while (!rtpe_shutdown) {
+		// see websocket_write_raw() for locking logic
+
+		mutex_lock(&websocket_service_lock);
 		lws_service(websocket_context, 100);
+		mutex_unlock(&websocket_service_lock);
+
+		mutex_lock(&websocket_callback_lock);
+		mutex_unlock(&websocket_callback_lock);
+	}
 
 	websocket_cleanup();
 }
@@ -1017,3 +1173,8 @@ void websocket_start(void) {
 		return;
 	thread_create_detach_prio(websocket_loop, NULL, rtpe_config.scheduling, rtpe_config.priority, "websocket");
 }
+void websocket_stop(void) {
+	if (!websocket_context)
+		return;
+	lws_cancel_service(websocket_context);
+}
