@@ -1,11 +1,40 @@
 #ifndef _SDP_H_
 #define _SDP_H_
 
+
 #include <glib.h>
 #include "str.h"
 #include "call.h"
 #include "media_socket.h"
 
+enum command_type {
+	CMD_ADD = 0,
+	CMD_REM,
+	CMD_SUBST,
+};
+
+/* */
+struct sdp_substitute_attr {
+	str * value_a; /* from */
+	str * value_b; /* to */
+};
+
+/* A structure for SDP arbitrary manipulations on all levels of SDP:
+ * session (global), media (audio/video). Works only on `a=` lines.
+ */
+struct sdp_manipulations_common {
+	GQueue add_commands_glob;
+	GQueue add_commands_audio;
+	GQueue add_commands_video;
+
+	GHashTable * rem_commands_glob;
+	GHashTable * rem_commands_audio;
+	GHashTable * rem_commands_video;
+
+	GQueue subst_commands_glob;
+	GQueue subst_commands_audio;
+	GQueue subst_commands_video;
+};
 
 struct ice_candidate;
 
@@ -27,6 +56,7 @@ void sdp_free(GQueue *sessions);
 int sdp_replace(struct sdp_chopper *, GQueue *, struct call_monologue *, struct sdp_ng_flags *);
 int sdp_is_duplicate(GQueue *sessions);
 int sdp_create(str *out, struct call_monologue *, struct sdp_ng_flags *flags);
+const char *sdp_get_sendrecv(struct call_media *media);
 
 int sdp_parse_candidate(struct ice_candidate *cand, const str *s); // returns -1, 0, 1
 
