@@ -4,8 +4,9 @@
 
 struct codec_def_s;
 struct packet_sequencer_s;
-typedef struct codec_def_s codec_def_t;
+typedef const struct codec_def_s codec_def_t;
 typedef struct packet_sequencer_s packet_sequencer_t;
+struct rtp_payload_type;
 
 enum media_type {
 	MT_UNKNOWN = 0,
@@ -18,6 +19,7 @@ enum media_type {
 
 
 #include "str.h"
+#include "rtplib.h"
 
 INLINE enum media_type codec_get_type(const str *type) {
 	if (!type || !type->len)
@@ -34,6 +36,13 @@ INLINE enum media_type codec_get_type(const str *type) {
 }
 
 
+// 0:  exact match
+// 1:  a is compatible with b (not necessarily the other way around)
+// -1: a is not compatible with b
+typedef int format_cmp_f(const struct rtp_payload_type *a, const struct rtp_payload_type *b);
+typedef bool format_print_f(GString *, const struct rtp_payload_type *);
+
+
 #ifndef WITHOUT_CODECLIB
 
 
@@ -46,6 +55,7 @@ INLINE enum media_type codec_get_type(const str *type) {
 #include <bcg729/encoder.h>
 #include <bcg729/decoder.h>
 #endif
+#include <opus.h>
 
 #define AMR_FT_TYPES 14
 
@@ -59,7 +69,9 @@ struct resample_s;
 struct seq_packet_s;
 struct rtp_payload_type;
 union codec_options_u;
+struct encoder_callback_s;
 struct dtx_method_s;
+struct fraction;
 
 typedef struct codec_type_s codec_type_t;
 typedef struct decoder_s decoder_t;
@@ -68,55 +80,61 @@ typedef struct format_s format_t;
 typedef struct resample_s resample_t;
 typedef struct seq_packet_s seq_packet_t;
 typedef union codec_options_u codec_options_t;
+typedef struct encoder_callback_s encoder_callback_t;
 typedef struct dtx_method_s dtx_method_t;
 
 typedef int packetizer_f(AVPacket *, GString *, str *, encoder_t *);
 typedef void format_init_f(struct rtp_payload_type *);
-typedef void set_enc_options_f(encoder_t *, const str *, const str *);
-typedef void set_dec_options_f(decoder_t *, const str *, const str *);
-typedef int format_cmp_f(const struct rtp_payload_type *, const struct rtp_payload_type *);
+typedef void set_enc_options_f(encoder_t *, const str *);
+typedef void set_dec_options_f(decoder_t *, const str *);
+typedef void select_encoder_format_f(encoder_t *, format_t *requested_format, const format_t *input_format,
+		const struct rtp_codec_format *fmtp);
+typedef void select_decoder_format_f(decoder_t *, const struct rtp_codec_format *fmtp);
+
+typedef int format_parse_f(struct rtp_codec_format *, const str *fmtp);
+typedef void format_answer_f(struct rtp_payload_type *, const struct rtp_payload_type *);
 
 
 
 struct codec_type_s {
-	void (*def_init)(codec_def_t *);
+	void (*def_init)(struct codec_def_s *);
 
-	const char *(*decoder_init)(decoder_t *, const str *, const str *);
+	const char *(*decoder_init)(decoder_t *, const str *);
 	int (*decoder_input)(decoder_t *, const str *data, GQueue *);
 	void (*decoder_close)(decoder_t *);
 
-	const char *(*encoder_init)(encoder_t *, const str *, const str *);
+	const char *(*encoder_init)(encoder_t *, const str *);
 	int (*encoder_input)(encoder_t *, AVFrame **);
 	void (*encoder_got_packet)(encoder_t *);
 	void (*encoder_close)(encoder_t *);
 };
 
-struct amr_cmr {
-	struct timeval cmr_in_ts;
-	unsigned int cmr_in;
+struct encoder_callback_s {
+	struct {
+		struct timeval cmr_in_ts;
+		unsigned int cmr_in;
 
-	struct timeval cmr_out_ts;
-	unsigned int cmr_out;
+		struct timeval cmr_out_ts;
+		unsigned int cmr_out;
+	} amr;
+
+	struct {
+		struct timeval cmr_in_ts;
+		unsigned int cmr_in;
+	} evs;
 };
-
 union codec_options_u {
 	struct {
-		int interleaving;
-		unsigned int mode_set; // bitfield
-		int mode_change_period;
-		int mode_change_interval;
-		unsigned int octet_aligned:1;
-		unsigned int crc:1;
-		unsigned int robust_sorting:1;
-		unsigned int mode_change_neighbor:1;
-
 		const unsigned int *bits_per_frame;
 		const unsigned int *bitrates;
 
-		struct amr_cmr cmr; // input from external calling code
-
+		int mode_change_interval;
 		int cmr_interval;
 	} amr;
+
+	struct {
+		enum evs_bw max_bw;
+	} evs;
 };
 
 enum dtx_method {
@@ -127,9 +145,26 @@ enum dtx_method {
 	NUM_DTX_METHODS
 };
 
+struct fraction {
+	int mult;
+	int div;
+};
+INLINE int fraction_mult(int a, const struct fraction *f) {
+	return a * f->mult / f->div;
+}
+INLINE int fraction_div(int a, const struct fraction *f) {
+	return a * f->div / f->mult;
+}
+INLINE long fraction_multl(long a, const struct fraction *f) {
+	return a * f->mult / f->div;
+}
+INLINE long fraction_divl(long a, const struct fraction *f) {
+	return a * f->div / f->mult;
+}
+
 struct codec_def_s {
 	const char * const rtpname;
-	int clockrate_mult;
+	struct fraction default_clockrate_fact;
 	const int avcodec_id;
 	const char * const avcodec_name_enc;
 	const char * const avcodec_name_dec;
@@ -137,9 +172,15 @@ struct codec_def_s {
 	int default_channels;
 	const int default_bitrate;
 	int default_ptime;
+	int minimum_ptime;
 	const char *default_fmtp;
+	format_parse_f * const format_parse;
 	format_cmp_f * const format_cmp;
+	format_print_f * const format_print;
+	format_answer_f * const format_answer;
 	packetizer_f * const packetizer;
+	select_encoder_format_f * const select_encoder_format;
+	select_decoder_format_f * const select_decoder_format;
 	const int bits_per_sample;
 	const enum media_type media_type;
 	const str silence_pattern;
@@ -164,8 +205,8 @@ struct codec_def_s {
 	const codec_type_t *codec_type;
 
 	// libavcodec
-	AVCodec *encoder;
-	AVCodec *decoder;
+	const AVCodec *encoder;
+	const AVCodec *decoder;
 };
 
 struct format_s {
@@ -182,6 +223,7 @@ struct resample_s {
 enum codec_event {
 	CE_AMR_CMR_RECV,
 	CE_AMR_SEND_CMR,
+	CE_EVS_CMR_RECV,
 };
 
 struct dtx_method_s {
@@ -200,8 +242,10 @@ struct dtx_method_s {
 };
 
 struct decoder_s {
-	const codec_def_t *def;
+	codec_def_t *def;
+	struct fraction clockrate_fact;
 	codec_options_t codec_options;
+	union codec_format_options format_options;
 	dtx_method_t dtx;
 
 	format_t in_format,
@@ -231,6 +275,8 @@ struct decoder_s {
 			unsigned int event;
 			unsigned long duration;
 		} dtmf;
+		void *evs;
+		OpusDecoder *opus;
 	} u;
 
 	unsigned long rtp_ts;
@@ -243,14 +289,18 @@ struct decoder_s {
 
 struct encoder_s {
 	format_t requested_format,
+		 input_format,
 		 actual_format;
 
-	const codec_def_t *def;
+	codec_def_t *def;
+	struct fraction clockrate_fact;
 	codec_options_t codec_options;
+	encoder_callback_t callback;
+	union codec_format_options format_options;
 
 	union {
 		struct {
-			AVCodec *codec;
+			const AVCodec *codec;
 			AVCodecContext *avcctx;
 
 			union {
@@ -265,6 +315,12 @@ struct encoder_s {
 #ifdef HAVE_BCG729
 		bcg729EncoderChannelContextStruct *bcg729;
 #endif
+		struct {
+			void *ctx;
+			void *ind_list;
+			struct timeval cmr_in_ts;
+		} evs;
+		OpusEncoder *opus;
 	} u;
 	AVPacket *avpkt;
 	AVAudioFifo *fifo;
@@ -298,15 +354,17 @@ void codeclib_init(int);
 void codeclib_free(void);
 
 
-const codec_def_t *codec_find(const str *name, enum media_type);
-const codec_def_t *codec_find_by_av(enum AVCodecID);
+codec_def_t *codec_find(const str *name, enum media_type);
+codec_def_t *codec_find_by_av(enum AVCodecID);
 
+int codec_parse_fmtp(codec_def_t *def, struct rtp_codec_format *fmtp, const str *fmtp_string,
+		union codec_format_options *copy);
 
-decoder_t *decoder_new_fmt(const codec_def_t *def, int clockrate, int channels, int ptime,
+decoder_t *decoder_new_fmt(codec_def_t *def, int clockrate, int channels, int ptime,
 		const format_t *resample_fmt);
-decoder_t *decoder_new_fmtp(const codec_def_t *def, int clockrate, int channels, int ptime,
+decoder_t *decoder_new_fmtp(codec_def_t *def, int clockrate, int channels, int ptime,
 		const format_t *resample_fmt,
-		const str *fmtp, const str *codec_opts);
+		struct rtp_codec_format *fmtp, const str *fmtp_string, const str *codec_opts);
 void decoder_close(decoder_t *dec);
 int decoder_input_data(decoder_t *dec, const str *data, unsigned long ts,
 		int (*callback)(decoder_t *, AVFrame *, void *u1, void *u2), void *u1, void *u2);
@@ -320,10 +378,12 @@ int decoder_dtx(decoder_t *dec, unsigned long ts, int ptime,
 
 
 encoder_t *encoder_new(void);
-int encoder_config(encoder_t *enc, const codec_def_t *def, int bitrate, int ptime,
+int encoder_config(encoder_t *enc, codec_def_t *def, int bitrate, int ptime,
 		const format_t *requested_format, format_t *actual_format);
-int encoder_config_fmtp(encoder_t *enc, const codec_def_t *def, int bitrate, int ptime,
-		const format_t *requested_format, format_t *actual_format, const str *fmtp, const str *codec_opts);
+int encoder_config_fmtp(encoder_t *enc, codec_def_t *def, int bitrate, int ptime,
+		const format_t *input_format,
+		const format_t *requested_format, format_t *actual_format,
+		struct rtp_codec_format *fmtp, const str *fmtp_string, const str *codec_opts);
 void encoder_close(encoder_t *);
 void encoder_free(encoder_t *);
 int encoder_input_data(encoder_t *enc, AVFrame *frame,
@@ -341,6 +401,12 @@ void *packet_sequencer_force_next_packet(packet_sequencer_t *ps);
 int packet_sequencer_insert(packet_sequencer_t *ps, seq_packet_t *);
 
 
+void frame_fill_tone_samples(enum AVSampleFormat fmt, void *samples, unsigned int offset, unsigned int num,
+		unsigned int freq, unsigned int volume, unsigned int sample_rate, unsigned int channels);
+void frame_fill_dtmf_samples(enum AVSampleFormat fmt, void *samples, unsigned int offset, unsigned int num,
+		unsigned int event, unsigned int volume, unsigned int sample_rate, unsigned int channels);
+
+
 #include "auxlib.h"
 
 
@@ -380,13 +446,13 @@ INLINE int decoder_event(decoder_t *dec, enum codec_event event, void *ptr) {
 
 #else
 
-typedef int format_cmp_f(const void *, const void *);
 
 // stubs
 struct codec_def_s {
 	int dtmf;
 	int supplemental;
 	format_cmp_f * const format_cmp;
+	format_print_f * const format_print;
 	const str silence_pattern;
 };
 struct packet_sequencer_s {
@@ -400,7 +466,7 @@ INLINE void codeclib_free(void) {
 	;
 }
 
-INLINE const codec_def_t *codec_find(const str *name, enum media_type type) {
+INLINE codec_def_t *codec_find(const str *name, enum media_type type) {
 	return NULL;
 }
 INLINE void packet_sequencer_destroy(packet_sequencer_t *p) {
