@@ -55,6 +55,7 @@ struct stream_s {
 	int fd;
 	handler_t handler;
 	unsigned int forwarding_on:1;
+	double start_time;
 };
 typedef struct stream_s stream_t;
 
@@ -100,6 +101,7 @@ struct tag_s {
 	unsigned long id;
 	char *name;
 	char *label;
+	char *metadata;
 };
 typedef struct tag_s tag_t;
 
@@ -114,6 +116,8 @@ struct metafile_s {
 	char *output_dest;
 	off_t pos;
 	unsigned long long db_id;
+	unsigned int db_streams;
+	double start_time;
 
 	GStringChunk *gsc; // XXX limit max size
 
@@ -146,9 +150,11 @@ struct output_s {
 		*file_name,
 		*filename; // path + filename + suffix
 	const char *file_format;
+	const char *kind; // "mixed" or "single"
 	unsigned long long db_id;
 	gboolean skip_filename_extension;
 	unsigned int channel_mult;
+	double start_time;
 
 	AVFormatContext *fmtctx;
 	AVStream *avst;
